'''Define the InstFitter class.'''


__author__ = 'Chao Wu'


from functools import partial
from collections.abc import Iterable
from copy import deepcopy
from math import ceil
import numpy as np
from multiprocessing import Pool
from ..io.inputs import read_measurements_from_file, read_initial_values
from ..io.results import InstFitResults, InstFitMCResults
from .inst_simulate import InstSimulator
from .fit import Fitter
from ..solver.nlpsolver import InstMFAModel
from ..utils.progress import Progress


class InstFitter(Fitter, InstSimulator):
    '''
    Estimated fluxes are in the unit of umol/gCDW/s if concentrations in the unit of 
    umol/gCDW and timepoints in the unit of s.
    '''
    
    def set_measured_MDVs(self, fragmentid, timepoints, means, sds):
        '''
        Set measured MDVs at various timepoints.
        
        Parameters
        ----------
        fragmentid: str
            Metabolite ID + '_' + atom NOs, e.g., 'Glu_12345'.
        timepoints: float or list of float
            Timepoint(s).
        means: array or list of array
            Mean of measured MDV(s). len(means) should be equal to len(timepoints).
        sds: array or list of array
            Standard deviation of measured MDV(s). len(sds) should be equal to len(timepoints).
        '''
        
        if not isinstance(timepoints, Iterable):
            timepoints = [timepoints]
            means = [means]
            sds = [sds]
            
        for timepoint, mean, sd in zip(timepoints, means, sds):
            self.model.measured_inst_MDVs.setdefault(fragmentid, {})[timepoint] = [
                np.array(mean), 
                np.array(sd)
            ]
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_measured_MDVs, {fragmentid: timepoints}))
            
    
    def set_measured_MDVs_from_file(self, file):
        '''
        Read measured MDVs at various timepoints from file.
        
        Parameters
        ----------
        file: file path
            Path of tsv or excel file with fields "fragment_ID", "time", "mean" and "sd".
            "fragment_ID" is metabolite ID + '_' + atom NOs, e.g., 'Glu_12345';
            "time" is timepoint when MDVs are measured (while some timepoints could be missing);
            "mean" and "sd" are the mean and standard deviation of MDV with element seperated by ','.

            Header line starts with "#", and will be skiped.
        '''
        
        measMDVs = read_measurements_from_file(file, inst_data = True)
        
        fragmentid_tpoints = {}
        for [emuid, timepoint], [mean, sd] in measMDVs.iterrows():
            timepoint = float(timepoint)
            self.model.measured_inst_MDVs.setdefault(emuid, {})[timepoint] = [
                np.array(list(map(float, mean.split(',')))), 
                np.array(list(map(float, sd.split(','))))
            ]
            fragmentid_tpoints.setdefault(emuid, []).append(timepoint)
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_measured_MDVs, fragmentid_tpoints))
    
    
    def _unset_measured_MDVs(self, fragmentid_tpoints):
        '''
        Parameters
        ----------
        fragmentid_tpoints: dict
            measured MDV ID => list of timepoints.
        '''
        
        for fragmentid, timepoints in fragmentid_tpoints.items():
            if fragmentid in self.model.measured_inst_MDVs:
                for timepoint in timepoints:
                    if timepoint in self.model.measured_inst_MDVs[fragmentid]:
                        self.model.measured_inst_MDVs[fragmentid].pop(timepoint)
                        
                if not self.model.measured_inst_MDVs[fragmentid]:
                    self.model.measured_inst_MDVs.pop(fragmentid)
                    
            
    def set_concentration_bounds(self, metabid, bounds):
        '''
        Set lower and upper bounds for concentration in unit of umol/gCDW.
        
        Parameters
        ----------
        metabid: str or 'all'
            Metabolite ID. If 'all', all concentrations will be set to the range.
        bounds: 2-list
            [lower bound, upper bound]. Lower bound is not allow to equal upper bound.
        '''
        
        bounds = list(map(float, bounds))

        metabids = []
        if bounds[0] < bounds[1]:   
            if metabid == 'all':
                for metabid in self.model.metabolites:
                    self.model.concentrations_bounds[metabid] = [max(0.0, bounds[0]), bounds[1]]
                    metabids.append(metabid)
            elif metabid in self.model.metabolites:
                self.model.concentrations_bounds[metabid] = [max(0.0, bounds[0]), bounds[1]]
                metabids = [metabid]
            else:
                raise ValueError(f'concentration range set to nonexistent metabolite {metabid}')
        else:
            raise ValueError('concentration lower bound should be less than upper bound')    
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_concentration_bounds, metabids))
            
            
    def _set_default_concentration_bounds(self):
        '''
        This method assign bounds of [0.01, 100] (umol/gCDW) for concentrations
        not set by set_concentration_bounds
        '''
        
        defBnds = [0.01, 100]
        
        metabids = []
        for metabid in self.model.metabolites:
            if metabid not in self.model.concentrations_bounds:
                self.model.concentrations_bounds[metabid] = defBnds
                metabids.append(metabid)
                
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_concentration_bounds, metabids))        
        
    
    def _unset_concentration_bounds(self, metabids):
        '''
        Parameters
        ----------
        metabids: str or list of str
            Metabolite ID(s).
        '''
        
        if not isinstance(metabids, Iterable):
            metabids = [metabids]
        
        for metabid in metabids:
            if metabid in self.model.concentrations_bounds:
                self.model.concentrations_bounds.pop(metabid)
                
                
    def _decompose_network(self, n_jobs):
        '''
        Parameters
        ----------
        n_jobs: int
            # of jobs to run in parallel.
        '''
        
        if not self.model.measured_inst_MDVs:
            raise ValueError('call set_measured_MDV or set_measured_MDVs_from_file first')
        
        if not self.model.EAMs:
            if n_jobs <= 0:
                raise ValueError('n_jobs should be a positive value')    
            else:
                self.model.target_EMUs = list(self.model.measured_inst_MDVs.keys())
                
                metabids = []
                atom_nos = []
                for emuid in self.model.target_EMUs:
                    metabid, atomNOs = emuid.split('_')
                    metabids.append(metabid)
                    atom_nos.append(atomNOs)
                
                EAMs = self.model._decompose_network(
                    metabids, atom_nos, 
                    lump = False, 
                    n_jobs = n_jobs
                )
                for size, EAM in EAMs.items():
                    self.model.EAMs[size] = EAM
                
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_decomposition)
            
    
    def _set_timepoints(self):
        
        if not self.model.timepoints:
            self.calculator._set_timepoints()
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_timepoints)
        
        
    def _calculate_matrix_Ms_derivatives_p(self):
        
        if not self.model.matrix_Ms_der_p:
            self.calculator._calculate_matrix_Ms_derivatives_p()
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_matrix_Ms_derivatives_p)
            
        
    def _unset_matrix_Ms_derivatives_p(self):
        
        self.model.matrix_Ms_der_p.clear()    
        
    
    def _calculate_measured_inst_MDVs_inversed_covariance_matrix(self):
        
        if not self.model.measured_inst_MDVs_inv_cov:
            self.calculator._calculate_measured_inst_MDVs_inversed_covariance_matrix()
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_measured_inst_MDVs_inversed_covariance_matrix)
            
            
    def _unset_measured_inst_MDVs_inversed_covariance_matrix(self):
        
        self.model.measured_inst_MDVs_inv_cov = None    
        
    
    def _calculate_initial_matrix_Xs_derivatives_p(self):
        
        if not self.model.initial_matrix_Xs_der_p:
            self.calculator._calculate_initial_matrix_Xs_derivatives_p()
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_initial_matrix_Xs_derivatives_p)
    
    
    def _unset_initial_matrix_Xs_derivatives_p(self):
        
        self.model.initial_matrix_Xs_der_p.clear()
    
    
    def _calculate_initial_matrix_Ys_derivatives_p(self):
        
        if not self.model.initial_matrix_Ys_der_p:
            self.calculator._calculate_initial_matrix_Ys_derivatives_p()
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_initial_matrix_Ys_derivatives_p)
    
    
    def _unset_initial_matrix_Ys_derivatives_p(self):
        
        self.model.initial_matrix_Ys_der_p.clear()
        
        
    def _estimate_concentrations_range(self):
        
        if not self.model.concentrations_range:
            for metabid in self.model.concids:
                self.model.concentrations_range[metabid] = self.model.concentrations_bounds[metabid]
                
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_concentrations_range, self.model.concids))
            
        
    def _unset_concentrations_range(self, metabids):
        '''
        Parameters
        ----------
        metabids: str or list of str
            Metabolite ID(s).
        '''
        
        if not isinstance(metabids, Iterable):
            metabids = [metabids]
        
        for metabid in metabids:
            if metabid in self.model.concentrations_range:
                self.model.concentrations_range.pop(metabid)
        
        
    def prepare(self, dilution_from = None, n_jobs = 1):
        '''
        Parameters
        ----------
        dilution_from: str or list of str
            ID(s) of unlabeled (inactive) metabolite leading to dilution effect.
            These metabolites have zero stoichiometric coefficients in reaction network.
        n_jobs: int
            If n_jobs > 1, decomposition job will run in parallel.
        '''    
        
        self._decompose_network(n_jobs)
        self._set_timepoints()
        self._calculate_null_space()
        self._calculate_transform_matrix()
        self._lambdify_matrix_As_and_Bs()
        self._calculate_matrix_As_and_Bs_derivatives_p('inst', n_jobs)
        self._lambdify_matrix_Ms()
        self._calculate_matrix_Ms_derivatives_p()
        self._calculate_substrate_MDVs(dilution_from)
        self._calculate_substrate_MDV_derivatives_p('inst', dilution_from)
        self._calculate_measured_inst_MDVs_inversed_covariance_matrix()
        self._calculate_measured_fluxes_inversed_covariance_matrix()
        self._calculate_measured_fluxes_derivative_p('inst')
        self._calculate_initial_matrix_Xs()
        self._calculate_initial_matrix_Ys()
        self._calculate_initial_matrix_Xs_derivatives_p()
        self._calculate_initial_matrix_Ys_derivatives_p()
        self._set_default_flux_bounds()
        self._estimate_fluxes_range(self.model.unbalanced_metabolites)
        self._set_default_concentration_bounds()
        self._estimate_concentrations_range()
    
        
    def _check_dependencies(self, fit_measured_fluxes):
        '''
        Parameters
        ----------
        fit_measured_fluxes: bool
            Whether to fit measured fluxes.
        '''

        if not self.model.net_fluxes_bounds:
            raise ValueError('call set_flux_bounds first')
        if not self.model.concentrations_bounds:
            raise ValueError('call set_concentration_bounds first')
        if not self.model.measured_inst_MDVs:
            raise ValueError('call call set_measured_MDV or set_measured_MDVs_from_file first')
        if not self.model.measured_fluxes:
            raise ValueError('call set_measured_flux or set_measured_fluxes_from_file first')    
        if not self.model.labeling_strategy:
            raise ValueError('call labeling_strategy first')
            
        checklist = [
            not self.model.target_EMUs, 
            self.model.transform_matrix is None, 
            self.model.null_space is None,
            self.model.measured_inst_MDVs_inv_cov is None, 
            not self.model.matrix_As, 
            not self.model.matrix_Bs,
            not self.model.matrix_Ms, 
            not self.model.matrix_As_der_p, 
            not self.model.matrix_Bs_der_p, 
            not self.model.matrix_Ms_der_p, 
            not self.model.substrate_MDVs, 
            not self.model.substrate_MDVs_der_p,
            self.model.measured_fluxes_der_p is None, 
            not self.model.initial_matrix_Xs, 
            not self.model.initial_matrix_Ys,
            not self.model.initial_matrix_Xs_der_p, 
            not self.model.initial_matrix_Ys_der_p, 
            not self.model.timepoints
        ]
        if fit_measured_fluxes:
            checklist.append(self.model.measured_fluxes_inv_cov is None)
        
        if any(checklist):
            raise ValueError('call prepare first')


    def solve(
            self, 
            fit_measured_fluxes = True, 
            ini_fluxes = None, 
            ini_concs = None, 
            solver = 'slsqp', 
            tol = 1e-6, 
            max_iters = 400, 
            show_progress = True
    ):
        '''
        Parameters
        ----------
        fit_measured_fluxes: bool
            Whether to fit measured fluxes.
        ini_fluxes: ser or file in .tsv or .xlsx
            Initial values of net fluxes.
        ini_concs: ser or file in .tsv or .xlsx
            Initial values of concentrations.
        solvor: {"slsqp", "ralg"}
            * If "slsqp", scipy.optimize.minimze will be used.
            * If "ralg", openopt NLP solver will be used.
        tol: float
            Tolerance for termination.
        max_iters: int
            Maximum # of iterations.
        show_progress: bool
            Whether to show the progress bar. 
        '''            
                
        self._check_dependencies(fit_measured_fluxes)

        if ini_fluxes is not None:
            iniFluxes = read_initial_values(ini_fluxes, self.model.netfluxids)
        else:
            iniFluxes = ini_fluxes

        if ini_concs is not None:
            iniConcs = read_initial_values(ini_concs, self.model.concids)
        else:
            iniConcs = ini_concs
        
        optModel = InstMFAModel(self.model, fit_measured_fluxes, solver)    
        optModel.build_objective()
        optModel.build_gradient()
        optModel.build_flux_and_conc_bound_constraints()
        optModel.build_initial_flux_and_conc_values(
            ini_netfluxes = iniFluxes, 
            ini_concs = iniConcs
        )
        
        with Progress('INST fitting', silent = not show_progress):
            res = optModel.solve_flux(tol, max_iters)

        return InstFitResults(
            *res[:8], 
            deepcopy(res[8]), 
            res[9], 
            deepcopy(res[10]), 
            *res[11:]
        )    


    def _solve_with_confidence_intervals(
            self, 
            fit_measured_fluxes, 
            ini_fluxes, 
            ini_concs, 
            solver, 
            tol, 
            max_iters, 
            nruns
    ):
        '''
        Parameters
        ----------
        fit_measured_fluxes: bool
            Whether to fit measured fluxes.
        ini_fluxes: ser or file in .tsv or .xlsx or None
            Initial values of net fluxes.
        ini_concs: ser or file in .tsv or .xlsx or None
            Initial values of concentrations.    
        solvor: {"slsqp", "ralg"}
            * If "slsqp", scipy.optimize.minimze will be used.
            * If "ralg", openopt NLP solver will be used.
        tol: float
            Tolerance for termination.
        max_iters: int
            Maximum # of iterations.
        nruns: int
            # of estimations in each worker.
        '''

        import platform
        if platform.system() == 'Linux':
            import os
            os.sched_setaffinity(os.getpid(), range(os.cpu_count()))

        self._lambdify_matrix_As_and_Bs()
        self._lambdify_matrix_Ms()
        
        if ini_fluxes is not None:
            iniFluxes = read_initial_values(ini_fluxes, self.model.netfluxids)
        else:
            iniFluxes = ini_fluxes

        # if ini_concs is not None:
        #     iniConcs = read_initial_values(ini_concs, self.model.concids)
        # else:
        #     iniConcs = ini_concs
        
        optTotalfluxesSet = []
        optNetfluxesSet = []
        optConcsSet = []
        for _ in range(nruns):
            self.calculator._generate_random_fluxes()
            self.calculator._generate_random_inst_MDVs()
            
            optModel = InstMFAModel(self.model, fit_measured_fluxes, solver)
            optModel.build_objective()
            optModel.build_gradient()
            optModel.build_flux_and_conc_bound_constraints()
            optModel.build_initial_flux_and_conc_values(ini_netfluxes = iniFluxes)
            
            try:
                while True:
                    (optTotalfluxes, 
                     optNetfluxes, 
                     optConcs, 
                      *_, 
                     isSuccess
                    ) = optModel.solve_flux(tol, max_iters)
                    if isSuccess:
                        break
            except:
                continue
            optTotalfluxesSet.append(optTotalfluxes)
            optNetfluxesSet.append(optNetfluxes)
            optConcsSet.append(optConcs)
            
            self.calculator._reset_measured_fluxes()
            self.calculator._reset_measured_inst_MDVs()
            
        return optTotalfluxesSet, optNetfluxesSet, optConcsSet


    def solve_with_confidence_intervals(
            self, 
            fit_measured_fluxes = True, 
            ini_fluxes = None, 
            ini_concs = None, 
            solver = 'slsqp', 
            tol = 1e-6, 
            max_iters = 400, 
            n_runs = 100, 
            n_jobs = 1, 
            show_progress = True
    ):
        '''
        Parameters
        ----------
        fit_measured_fluxes: bool
            Whether to fit measured fluxes.
        ini_fluxes: ser or file in .tsv or .xlsx
            Initial values of net fluxes.
        ini_concs: ser or file in .tsv or .xlsx
            Initial values of concentrations.   
        solvor: {"slsqp", "ralg"}
            * If "slsqp", scipy.optimize.minimze will be used.
            * If "ralg", openopt NLP solver will be used.
        tol: float
            Tolerance for termination.
        max_iters: int
            Max # of iterations.
        show_progress: bool
            Whether to show the progress bar.
        n_runs: int
            # of runs to estimate confidence intervals.
        n_jobs: int
            # of jobs to run in parallel.
        '''
        
        self._check_dependencies(fit_measured_fluxes)
        
        self._unset_matrix_As_and_Bs()
        self._unset_matrix_Ms()

        if n_runs <= n_jobs:
            nruns_worker = 1
        else:
            nruns_worker = ceil(n_runs/n_jobs)
        
        pool = Pool(processes = n_jobs)
        with Progress('INST fitting with CIs', silent = not show_progress):
            resSet = []
            for _ in range(n_jobs):
                res = pool.apply_async(
                    func = self._solve_with_confidence_intervals, 
                    args = (
                        fit_measured_fluxes,
                        ini_fluxes,
                        ini_concs,
                        solver,
                        tol,
                        max_iters,
                        nruns_worker
                    )
                )
                resSet.append(res)
            
            pool.close()    
            pool.join()
        
        resSet = [res.get() for res in resSet]
        
        totalFluxesSet = []
        netFluxesSet = []
        concsSet = []
        for totalFluxesSubset, netFluxesSubset, concsSubset in resSet:
            totalFluxesSet.extend(totalFluxesSubset)
            netFluxesSet.extend(netFluxesSubset)
            concsSet.extend(concsSubset)
        
        self._lambdify_matrix_As_and_Bs()
        self._lambdify_matrix_Ms()

        return InstFitMCResults(totalFluxesSet, netFluxesSet, concsSet)     
        


'''Define data input functions.'''


__author__ = 'Chao Wu'


from os.path import splitext, isfile
import re
import pandas as pd


def read_model_from_file(file):
    '''
    Parameters
    ----------
    file: file path
        tsv or excel file with reactions.
    '''
    
    ext = splitext(file)[1]
    if re.search(r'tsv', ext):
        data = pd.read_csv(
            file, 
            sep = '\t', 
            comment = '#', 
            header = None, 
            names = ['subs', 'pros', 'rev'], 
            index_col = 0
        )
    elif re.search(r'xls', ext):
        data = pd.read_excel(
            file, 
            comment = '#', 
            header = None, 
            names = ['subs', 'pros', 'rev'], 
            index_col = 0
        )
    else:
        raise TypeError('can only read from .tsv or excel file')
    
    data = data.dropna()
    
    return data


def read_preset_values_from_file(file):
    '''
    Parameters
    ----------
    file: file path
        tsv or excel file.
    '''
    
    ext = splitext(file)[1]
    if re.search(r'tsv', ext):
        data = pd.read_csv(
            file, 
            sep = '\t', 
            comment = '#', 
            header = None, 
            names = ['value'], 
            index_col = 0, 
        ).squeeze()
    elif re.search(r'xls', ext):
        data = pd.read_excel(
            file, 
            comment = '#', 
            header = None, 
            names = ['value'], 
            index_col = 0, 
        ).squeeze()
    else:
        raise TypeError('can only read from .tsv or excel file')
    
    data = data.dropna()

    return data
        

def read_initial_values(ini, ids):
    '''
    Parameters
    ----------
    ini: ser of file in .tsv or .xlsx
        Initial values.
    ids: list
        IDs of fluxes or concentrations in correct order.
    '''

    if isfile(ini):
        ini = read_preset_values_from_file(ini)
        ini = ini[ids]
    elif isinstance(ini, pd.Series):
        ini = ini[ids]
    else:
        raise ValueError('initial values should be in pd.Series or file')

    return ini    

        
def read_measurements_from_file(file, inst_data = False):
    '''
    Parameters
    ----------
    file: file path
        tsv or excel file.
    inst_data: bool
        If True, additional column of timepoints is used as index (multiindex).
    '''
    
    if inst_data:
        indexCols = [0, 1]
    else:
        indexCols = 0
    
    ext = splitext(file)[1]
    if re.search(r'tsv', ext):
        data = pd.read_csv(
            file, 
            sep = '\t', 
            comment = '#', 
            header = None, 
            names = ['mean', 'sd'], 
            index_col = indexCols
        )                        
    elif re.search(r'xls', ext):
        data = pd.read_excel(
            file, 
            comment = '#', 
            header = None
        )
        data.set_index(indexCols, inplace = True)
        data.columns = ['mean', 'sd']
    else:
        raise TypeError('can only read from .tsv or excel file')
    
    data = data.dropna()
        
    return data



'''Patch OpenOpt.'''


__author__ = 'Chao Wu'


from importlib.util import find_spec
from pathlib import Path
import re


OPENOPT_MODULES = ['ooIter', 'result', 'runProbSolver']


def _patch_module(spec_dir, mod):
    mod_file = Path(spec_dir) / 'kernel' / f'{mod}.py'
    
    with open(mod_file) as f:
        old_code = f.read()
    
    new_code = re.sub(r'\bclock\b', 'perf_counter', old_code)
    
    if new_code != old_code:
        with open(mod_file, 'w') as f:
            f.write(new_code)
        
        print(f'{mod_file} patched.')
    

def apply_patch():
    spec_dir = find_spec('openopt').submodule_search_locations[0]
    
    for mod in OPENOPT_MODULES:
        _patch_module(spec_dir, mod)


'''Define the Optimizer class.'''


__author__ = 'Chao Wu'


from functools import partial
from collections.abc import Iterable
from ..solver.lpsolver import FBAModel
from ..io.results import FBAResults, FVAResults
from ..utils.progress import Progress
from ..utils.context import Context


class Optimizer():
    '''
    Parameters
    ----------
    model: Model
        Freeflux Model.
    '''
    
    def __init__(self, model):
        '''
        Parameters
        ----------
        model: Model
            Freeflux Model.
        '''
        
        self.model = model
        self.contexts = []   
        
    
    def __enter__(self):
        
        self.contexts.append(Context())

        return self
        
        
    def __exit__(self, type, value, traceback):
        
        context = self.contexts.pop()
        context.undo()
        
    
    def _set_flux_bounds(self, fluxid, bounds):
        '''
        Parameters
        ----------
        fluxid: str
            Flux ID, i.e. reaction ID. For irreversible reaction, the lower bound of range will 
            be set to zero ignorant of bounds[0].
        bounds: 2-list
            [lower bound, upper bound].
        '''
        
        bounds = list(map(float, bounds))

        if self.model.reactions_info[fluxid].reversible:
            self.model.net_fluxes_bounds[fluxid] = list(bounds)
        else:
            if bounds[1] >= 0.0:
                self.model.net_fluxes_bounds[fluxid] = [max(0.0, bounds[0]), bounds[1]]
            else:
                raise ValueError(f'flux bounds of {fluxid} conflict with its reversibility')
        
    
    def set_flux_bounds(self, fluxid, bounds):
        '''
        Set lower and upper bounds for metabolic flux. If called several times, 
        the lastest setting overrides previous ones if conflicted.
        
        Parameters
        ----------
        fluxid: str or 'all'
            Flux ID, i.e. reaction ID. This method is used to set the range of net flux.
            For irreversible reaction, the lower bound will be set to zero ignorant of bounds[0].
            
            If 'all', all fluxes will be set to the range.
        bounds: 2-list
            [lower bound, upper bound]. Let lower bound equals upper bound to set fixed value.
        '''
        
        fluxids = []
        if bounds[0] <= bounds[1]:
            if fluxid == 'all':
                for fluxid in self.model.netfluxids:
                    self._set_flux_bounds(fluxid, bounds)
                    fluxids.append(fluxid)    
            elif fluxid in self.model.netfluxids:
                self._set_flux_bounds(fluxid, bounds)
                fluxids = [fluxid]
            else:
                raise ValueError(f'flux range set to nonexistent reaction {fluxid}')
        else:
            raise ValueError('lower bound of flux should not be greater than upper bound')
    
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_flux_bounds, fluxids))
    
                
    def _set_default_flux_bounds(self):
        '''
        This method assigns bounds of [-100, 100] for reversible reactions and 
        [0, 100] for irreversible reactions not yet set by set_flux_bounds
        '''
        
        defBunds = [-100, 100]
        
        fluxids = []
        for fluxid in self.model.netfluxids:
            if fluxid not in self.model.net_fluxes_bounds:
                self._set_flux_bounds(fluxid, defBunds)
                fluxids.append(fluxid)
                
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_flux_bounds, fluxids))

    
    def _unset_flux_bounds(self, fluxids):
        '''
        Parameters
        ----------
        fluxids: str or list of str
            Fluxe ID(s).
        '''
        
        if not isinstance(fluxids, Iterable):
            fluxids = [fluxids]
        
        for fluxid in fluxids:
            if fluxid in self.model.net_fluxes_bounds:
                self.model.net_fluxes_bounds.pop(fluxid)    


    def prepare(self):
        '''
        This method do some preparation work.
        '''
        
        self._set_default_flux_bounds()
        
        
    def _check_stoichiometric_matrix(self, exclude_metabs):
        '''
        Check and delete zero rows or columns usually caused by dilution reactions.

        Parameters
        ----------
        exclude_metabs: list
            Metabolite IDs, i.e., metabolites excluded from mass balance.
        '''
        
        netfluxids = self.model.netfluxids.copy()
        
        net_fluxes_bounds = self.model.net_fluxes_bounds.copy()
        
        stoy_mat = self.model.get_net_stoichiometric_matrix(exclude_metabs).copy(deep = 'all')
        
        maskZeroRows = (stoy_mat.T != 0.0).any()
        stoy_mat = stoy_mat[maskZeroRows]   
        
        maskZeroCols = (stoy_mat != 0.0).any()
        netfluxids_todrop = stoy_mat.columns[~maskZeroCols].tolist()

        if netfluxids_todrop:
            for fluxid in netfluxids_todrop:
                print(
                    f'{fluxid} not connected with other reactions, '
                    f'removed from flux balance'
                )

                if fluxid in netfluxids:
                    netfluxids.remove(fluxid)
                if fluxid in net_fluxes_bounds:
                    net_fluxes_bounds.pop(fluxid)

            stoy_mat = stoy_mat.T[maskZeroCols].T   

        return netfluxids, net_fluxes_bounds, stoy_mat

    
    def optimize(
            self, 
            objective, 
            direction = 'max', 
            exclude_metabs = None, 
            show_progress = True
    ):
        '''
        This method performs flux balance analysis by maximizing the objective.
        
        Parameters
        ----------
        objective: dict
            Reaction ID => coefficient, i.e., objective function.
        direction: {"max", "min"}
            Optimization direction.    
        exclude_metabs: list
            Metabolite IDs, i.e., metabolites excluded from mass balance.
        show_progress: bool
            Whether to show the progress bar.
            
        Returns
        -------
        FBAResults: FBAResults
        '''
        
        (netfluxids, 
         net_fluxes_bounds, 
         stoy_mat
        ) = self._check_stoichiometric_matrix(exclude_metabs)

        fbaModel = FBAModel()
        fbaModel.build_flux_variables(netfluxids, net_fluxes_bounds)
        fbaModel.build_objective(objective, direction)
        fbaModel.build_mass_balance_constraints(stoy_mat)
        
        with Progress('optimizing', silent = not show_progress):
            optObj, optFluxes = fbaModel.solve_flux()

        return FBAResults(objective, optObj, optFluxes)
        
        
    def estimate_fluxes_range(
            self, 
            objective = None, 
            gamma = 0, 
            exclude_metabs = None, 
            show_progress = True
    ):
        '''
        This method performs flux variability analysis and estimates the ranges of metabolic fluxes.
        If an objective is provided, it will be optimized in maximizing direction.
        
        Parameters
        ----------
        objective: dict
            Reaction ID => coefficient, i.e., objective function.
        gamma: float
            During evaluation of each flux, objective value is required to be 
            >= gamma*max(objective) with gamma in [0, 1]
        exclude_metabs: list
            Metabolite IDs, i.e., metabolites excluded from mass balance.
        show_progress: bool
            Whether to show the progress bar.    
            
        Returns
        -------
        FVAResults: FVAResults
        '''
        
        (netfluxids, 
         net_fluxes_bounds, 
         stoy_mat
        ) = self._check_stoichiometric_matrix(exclude_metabs)

        fvaModel = FBAModel()
        fvaModel.build_flux_variables(netfluxids, net_fluxes_bounds)
        fvaModel.build_mass_balance_constraints(stoy_mat)
        if objective is not None and gamma != 0:
            fvaModel.build_objective(objective, 'max')
            optObj, _ = fvaModel.solve_flux()
            fvaModel.remove_objective()
            fvaModel.build_objective_constraint(objective, optObj, gamma)
        
        with Progress('estimating flux ranges', silent = not show_progress):
        
            fluxRanges = {}
            for fluxid in self.model.netfluxids:
                if fluxid in netfluxids:
                    objective = {fluxid: 1}
                    
                    fvaModel.build_objective(objective, 'min')
                    fluxLB, _ = fvaModel.solve_flux()
                    fvaModel.remove_objective()
                    
                    fvaModel.build_objective(objective, 'max')
                    fluxUB, _ = fvaModel.solve_flux()
                    fvaModel.remove_objective()
                else:
                    fluxLB, fluxUB = self.model.net_fluxes_bounds[fluxid]
                
                fluxRanges[fluxid] = [fluxLB, fluxUB]

        return FVAResults(fluxRanges)
    


__author__ = 'Chao Wu'
__version__ = '0.3.7'


from importlib.util import find_spec
from .utils.patch import apply_patch
if find_spec('openopt') is not None:
    apply_patch()

from .core.metabolite import Metabolite
from .core.reaction import Reaction
from .core.model import Model
from .core.emu import EMU
from .core.mdv import MDV, get_natural_MDV, get_substrate_MDV, conv



'''Define the Progress class.'''


__author__ = 'Chao Wu'


from time import time, sleep
from datetime import timedelta
from threading import Thread


class Progress():
    '''
    Progress instances can be either used in a WITH statement or assigned to a variable. 
    In the second case, the start and stop method should be called explicitely.
    '''
    
    def __init__(self, descp, silent = False):
        '''
        Parameters
        ----------
        descp: str
            Description.
        silent: bool
            Whether to show progress bar.
        '''
        
        self.descp = descp
        self.count = 0
        self.is_running = False
        self.silent = silent
        
    
    def __enter__(self):
        
        self.start()
        
        
    def __exit__(self, type, value, traceback):
        
        self.stop()
    
        
    def _sleeptime(self):
        
        if self.descp == 'optimization':
            sleeptime = 0.1    
        elif self.descp == 'fitting with CIs':
            sleeptime = 10
        else:
            sleeptime = 2
            
        return sleeptime    
    
        
    def _show_progress(self, descp):
        
        t1 = time()
        while self.is_running:
            sleep(1)
            t2 = time()
            elapsed = timedelta(seconds = round(t2 - t1))
            print(f'{descp} [elapsed: {elapsed}]', end = '\r')
        print('')
        
        
    def start(self):
        
        if not self.silent:
            print('')
            self.is_running = True
            self.thread = Thread(target = self._show_progress, args = (self.descp,))
            self.thread.start()
        else:
            pass

    
    def stop(self):
        
        if not self.silent:
            self.is_running = False
            self.thread.join()
        else:
            pass
        
        


'''Define the FBAModel class.'''


__author__ = 'Chao Wu'


from pyomo.environ import (ConcreteModel, Var, Objective, Constraint, 
                           SolverFactory, maximize, minimize, value)
                                                   
                           
class FBAModel():
    
    def __init__(self):
        
        self.model = ConcreteModel()
        
        
    def build_flux_variables(self, fluxids, flux_bounds):
        '''
        Parameters
        ----------
        fluxids: list
            Flux IDs.
        flux_bounds: dict
            Reaction ID => [lower bound, upper bound].
        '''
        
        if set(fluxids) != flux_bounds.keys():
            raise ValueError('flux IDs not consistent with those with bounds')

        self.fluxids = fluxids
        
        def flux_bounds_rule(model, fluxid):
            return flux_bounds[fluxid]
            
        self.model.fluxes = Var(self.fluxids, bounds = flux_bounds_rule)

    
    def build_objective(self, objective, direction):
        '''
        Parameters
        ----------
        objective: dict
            Reaction ID => coefficient, i.e., objective function.
        direction: {"max", "min"}
            Optimization direction.
        '''
        
        if direction == 'max':
            sense = maximize
        elif direction == 'min':
            sense = minimize
        
        def obj_rule(model):
            objExpr = [coe*model.fluxes[rxnid] for rxnid, coe in objective.items()]
            return sum(objExpr)
        
        self.model.obj = Objective(rule = obj_rule, sense = sense)
        
        
    def build_mass_balance_constraints(self, stoy_mat):
        '''
        Parameters
        ----------
        stoy_mat: df
            Stoichiometric matrix.
        '''
        
        def mb_rule(model, metabid):
            fluxesExpr = [stoy_mat.loc[metabid, rxnid]*model.fluxes[rxnid] 
                          for rxnid in self.fluxids]
            return sum(fluxesExpr) == 0
            
        self.model.MBcstrs = Constraint(stoy_mat.index.tolist(), rule = mb_rule)
        
        
    def build_objective_constraint(self, objective, max_obj, gamma):
        '''
        Parameters
        ----------
        objective: dict
            Reaction ID => coefficient, i.e., objective function.
        max_obj: float
            Optimal objective.
        gamma: float
            A value in [0, 1].
        '''
        
        def objcstr_rule(model):
            objcstrExpr = [coe*model.fluxes[rxnid] for rxnid, coe in objective.items()]
            return sum(objcstrExpr) >= gamma*max_obj
            
        self.model.OBJcstr = Constraint(rule = objcstr_rule)
        
    
    def remove_objective(self):
        
        self.model.del_component(self.model.obj)
        
        
    def solve_flux(self):
        
        solver = SolverFactory('glpk')
        solver.solve(self.model)
        
        optObj = value(self.model.obj)
        
        optFluxes = {}
        for fluxid in self.fluxids:
            try:
                flux = value(self.model.fluxes[fluxid])
            except ValueError:
                print(f'flux value of {fluxid} not available')
                continue
            optFluxes[fluxid] = flux
        
        return optObj, optFluxes
        


'''Define the MFAModel and InstMFAModel class.'''


__author__ = 'Chao Wu'


import numpy as np
import pandas as pd
from scipy.linalg import pinv2
from scipy.optimize import minimize
try:
    from openopt import NLP
except ModuleNotFoundError:
    OPENOPT_INSTALLED = False
else:
    OPENOPT_INSTALLED = True
from ..utils.utils import Calculator
import warnings
warnings.filterwarnings('ignore', category = RuntimeWarning)


class MFAModel():
    '''
    Parameters
    ----------
    model: Model
        Freeflux Model.
    fit_measured_fluxes: bool
        Whether to fit measured fluxes.
    solvor: {"slsqp", "ralg"}
        * If "slsqp", scipy.optimize.minimze will be used.
        * If "ralg", openopt NLP solver will be used.
    '''
    
    def __init__(self, model, fit_measured_fluxes, solver = 'slsqp'):
        '''
        Parameters
        ----------
        model: Model
            Freeflux Model.
        fit_measured_fluxes: bool
            Whether to fit measured fluxes.
        solvor: {"slsqp", "ralg"}
            * If "slsqp", scipy.optimize.minimze will be used.
            * If "ralg", openopt NLP solver will be used.
        '''
        
        self.model = model
        self.calculator = Calculator(self.model)
        self.fit_measured_fluxes = fit_measured_fluxes
        self.solver = solver
        
        self.N = self.model.null_space
        self.T = self.model.transform_matrix
        
        self.ntotalfluxes = len(self.model.totalfluxids)
        
    
    def _calculate_difference_sim_exp_MDVs(self):
        
        simMDVs = self.calculator._calculate_MDVs()
        expMDVs = self.model.measured_MDVs
        diff = np.concatenate([simMDVs[emuid] - expMDVs[emuid][0] 
                               for emuid in self.model.target_EMUs])
        
        return diff
            
            
    def _calculate_difference_sim_exp_fluxes(self):
        
        simFluxes = self.model.total_fluxes[self.model.measured_fluxes.keys()]
        expFluxes = np.array([mean for mean, _ in self.model.measured_fluxes.values()])
        diff = simFluxes - expFluxes
        
        return diff
        
        
    def _calculate_sim_MDVs_derivative(self):
        
        simMDVs, simMDVsDer = self.calculator._calculate_MDVs_and_derivatives_p()
        expMDVs = self.model.measured_MDVs
        diff = np.concatenate([simMDVs[emuid] - expMDVs[emuid][0] 
                               for emuid in self.model.target_EMUs])
        
        dxsim_dp = np.vstack([simMDVsDer[emuid] for emuid in self.model.target_EMUs])
        
        return diff, dxsim_dp
        
    
    def _calculate_sim_fluxes_derivative(self):
        
        dvsim_dp = self.model.measured_fluxes_der_p
        
        return dvsim_dp


    def build_objective(self):
        
        def _f(u):
            self.model.total_fluxes[:] = self.N@u
            
            MDV_diff = self._calculate_difference_sim_exp_MDVs()
            MDV_inv_cov = self.model.measured_MDVs_inv_cov
            obj = MDV_diff@MDV_inv_cov@MDV_diff
            
            return obj
        
        def f1(u):
            return _f(u)
            
        def f2(u):
            obj1 = _f(u)
            
            flux_diff = self._calculate_difference_sim_exp_fluxes()
            flux_inv_cov = self.model.measured_fluxes_inv_cov
            obj2 = flux_diff@flux_inv_cov@flux_diff
            
            return obj1 + obj2
            
        self.f = f2 if self.fit_measured_fluxes else f1
        
    
    def build_gradient(self):
        
        def _df(u):
            self.model.total_fluxes[:] = self.N@u
            
            MDV_diff, MDV_der = self._calculate_sim_MDVs_derivative()
            MDV_inv_cov = self.model.measured_MDVs_inv_cov
            grad = MDV_der.T@MDV_inv_cov@MDV_diff
            
            return grad
            
        def df1(u):
            return _df(u)
            
        def df2(u):
            grad1 = _df(u)
            
            flux_der = self._calculate_sim_fluxes_derivative()
            flux_diff = self._calculate_difference_sim_exp_fluxes()
            flux_inv_cov = self.model.measured_fluxes_inv_cov
            grad2 = flux_der.T@flux_inv_cov@flux_diff
            
            return grad1 + grad2
        
        self.df = df2 if self.fit_measured_fluxes else df1

    
    def build_hessian(self):
        
        def _ddf(u):
            self.model.total_fluxes[:] = self.N@u
            
            _, MDV_der = self._calculate_sim_MDVs_derivative()
            MDV_inv_cov = self.model.measured_MDVs_inv_cov
            hess = MDV_der.T@MDV_inv_cov@MDV_der
            
            return hess
            
        def ddf1(u):
            return _ddf(u)
        
        def ddf2(u):
            hess1 = _ddf(u)
            
            flux_der = self._calculate_sim_fluxes_derivative()
            flux_inv_cov = self.model.measured_fluxes_inv_cov
            hess2 = flux_der.T@flux_inv_cov@flux_der
            
            return hess1 + hess2
        
        self.ddf = ddf2 if self.fit_measured_fluxes else ddf1    
    
            
    def build_flux_bound_constraints(self):
        
        A1 = self.N
        A2 = self.T@self.N
        A3 = -A2
        
        b1 = np.zeros(self.ntotalfluxes)
        vnet_lb, vnet_ub = np.array(list(self.model.net_fluxes_range.values())).T
        b2 = vnet_lb
        b3 = -vnet_ub
        
        A = np.vstack((A1, A2, A3))
        b = np.concatenate((b1, b2, b3))

        if self.solver == 'slsqp':
            self.constrs = {'type': 'ineq', 'fun': lambda u: A@u - b}
        elif self.solver == 'ralg':
            self.A = -A
            self.b = -b
        
        
    def build_initial_flux_values(self, ini_netfluxes = None):
        '''
        Parameters
        ----------
        ini_netfluxes: array
            Initial guess of net fluxes.
        '''
        
        if ini_netfluxes is None:
            vnet_lb, vnet_ub = np.array(list(self.model.net_fluxes_range.values())).T
            vnet_ini = np.random.uniform(low = vnet_lb, high = vnet_ub)
        else:
            vnet_ini = ini_netfluxes
            
        u_ini = pinv2(self.T@self.N)@vnet_ini
        
        self.x0 = u_ini
        
        
    def _initialize_total_fluxes(self):
        
        for fluxid in self.model.totalfluxids:
            self.model.total_fluxes[fluxid] = 0.0
        

    def _solve_flux_slsqp(self, tol, max_iters, disp):
        
        res = minimize(
            fun = self.f,
            x0 = self.x0,
            method = 'SLSQP',
            jac = self.df,
            constraints = self.constrs,
            options = {
                'ftol': tol,
                'maxiter': max_iters, 
                'disp': disp
            }
        )
       
        return res.fun, res.x, res.status in [0, 2]


    def _solve_flux_ralg(self, tol, max_iters, disp):

        if OPENOPT_INSTALLED:
            res = res = NLP(
                f = self.f, 
                x0 = self.x0, 
                df = self.df, 
                A = self.A, 
                b = self.b, 
                xtol = tol, 
                ftol = tol, 
                maxIter = max_iters, 
                iprint = 1 if disp else -1
            ).solve('ralg')

            return res.ff, res.xf, res.istop > 0 or res.istop == -7
        
        else:
            raise ModuleNotFoundError('install openopt first')


    def _calculate_residuals(self):
        
        MDV_diff = self._calculate_difference_sim_exp_MDVs()
        MDV_diag = np.diag(self.model.measured_MDVs_inv_cov)**0.5
        resids1 = MDV_diff*MDV_diag
        
        if self.fit_measured_fluxes:    
            flux_diff = self._calculate_difference_sim_exp_fluxes()
            flux_diag = np.diag(self.model.measured_fluxes_inv_cov)**0.5
            resids2 = flux_diff*flux_diag
            resids = np.concatenate((resids1, resids2))
        
        return resids         


    def _get_exp_and_sim_MDVs(self):

        exp_MDVs = self.model.measured_MDVs
        
        sim_MDVs_all = self.calculator._calculate_MDVs()
        sim_MDVs = {emuid: sim_MDVs_all[emuid] for emuid in self.model.target_EMUs}

        return exp_MDVs, sim_MDVs


    def _get_exp_and_sim_fluxes(self):

        if self.fit_measured_fluxes:
            exp_fluxes = self.model.measured_fluxes
            sim_fluxes = self.model.total_fluxes[self.model.measured_fluxes.keys()].to_dict()
        else:
            exp_fluxes = {}
            sim_fluxes = {}            

        return exp_fluxes, sim_fluxes


    def _get_nmeasurements(self, opt_resids):
        
        return opt_resids.size

         
    def _get_nparameters(self, opt_p):

        return opt_p.size


    def _get_hessian(self, opt_p):

        self.build_hessian()
        hess = self.ddf(opt_p)

        return hess


    def solve_flux(self, tol = 1e-6, max_iters = 400, disp = False):    
        
        self._initialize_total_fluxes()
        
        if self.solver == 'slsqp':
            opt_obj, opt_u, is_success = self._solve_flux_slsqp(tol, max_iters, disp)
        elif self.solver == 'ralg':
            opt_obj, opt_u, is_success = self._solve_flux_ralg(tol, max_iters, disp)
        else:
            raise ValueError('currently only "slsqp" and "ralg" are acceptable')
        
        opt_totalfluxes = pd.Series(self.N@opt_u, index = self.model.totalfluxids)
        opt_netfluxes = pd.Series(self.T@self.N@opt_u, index = self.model.netfluxids)
        
        opt_resids = self._calculate_residuals()
        
        exp_MDVs, sim_MDVs = self._get_exp_and_sim_MDVs()
        exp_fluxes, sim_fluxes = self._get_exp_and_sim_fluxes()
        
        nmeas = self._get_nmeasurements(opt_resids)
        nparams = self._get_nparameters(opt_u)
        
        hess = self._get_hessian(opt_u)

        _, dxsim_du = self._calculate_sim_MDVs_derivative()
        dvsim_du = self._calculate_sim_fluxes_derivative()
        
        return (
            opt_totalfluxes, 
            opt_netfluxes, 
            opt_obj, 
            opt_resids, 
            nmeas, 
            nparams, 
            sim_MDVs, 
            exp_MDVs, 
            sim_fluxes, 
            exp_fluxes, 
            hess, 
            self.N, 
            self.T, 
            dxsim_du, 
            dvsim_du, 
            self.model.measured_MDVs_inv_cov, 
            self.model.measured_fluxes_inv_cov, 
            is_success
        )
        
        
    
        
class InstMFAModel(MFAModel):

    def __init__(self, *args):
        
        super().__init__(*args)
        
        self.nfreefluxes = self.N.shape[1]
        self.nconcs = len(self.model.concids)
        self.nnetfluxes = len(self.model.netfluxids)
        
    
    def _calculate_difference_sim_exp_MDVs(self):
        
        simMDVs = self.calculator._calculate_inst_MDVs()
        expMDVs = self.model.measured_inst_MDVs
        diff = np.concatenate([simMDVs[emuid][t] - expMDVs[emuid][t][0] 
                               for emuid in self.model.target_EMUs
                               for t in expMDVs[emuid] if t != 0])
        
        return diff
        
        
    def _calculate_sim_MDVs_derivative(self):
        
        simMDVs, simMDVsDer = self.calculator._calculate_inst_MDVs_and_derivatives_p()
        expMDVs = self.model.measured_inst_MDVs
        diff = np.concatenate([simMDVs[emuid][t] - expMDVs[emuid][t][0] 
                               for emuid in self.model.target_EMUs
                               for t in expMDVs[emuid] if t != 0])
        
        dxsim_dp = np.vstack([simMDVsDer[emuid][t] 
                              for emuid in self.model.target_EMUs
                              for t in expMDVs[emuid] if t != 0])
        
        return diff, dxsim_dp
        
    
    def build_objective(self):
        
        def _f(p):
            u, c = p[:self.nfreefluxes], p[self.nfreefluxes:]
            self.model.total_fluxes[:] = self.N@u
            self.model.concentrations[:] = c
            
            MDV_diff = self._calculate_difference_sim_exp_MDVs()
            MDV_inv_cov = self.model.measured_inst_MDVs_inv_cov
            
            obj = MDV_diff@MDV_inv_cov@MDV_diff
            
            return obj
        
        def f1(p):
            return _f(p)
            
        def f2(p):
            obj1 = _f(p)
            
            flux_diff = self._calculate_difference_sim_exp_fluxes()
            flux_inv_cov = self.model.measured_fluxes_inv_cov
            obj2 = flux_diff@flux_inv_cov@flux_diff
            
            return obj1 + obj2
            
        self.f = f2 if self.fit_measured_fluxes else f1
    
        
    def build_gradient(self):
        
        def _df(p):
            u, c = p[:self.nfreefluxes], p[self.nfreefluxes:]
            self.model.total_fluxes[:] = self.N@u
            self.model.concentrations[:] = c
            
            MDV_diff, MDV_der = self._calculate_sim_MDVs_derivative()
            MDV_inv_cov = self.model.measured_inst_MDVs_inv_cov
            grad = MDV_der.T@MDV_inv_cov@MDV_diff
            
            return grad
            
        def df1(p):
            return _df(p)
            
        def df2(p):
            grad1 = _df(p)
            
            flux_der = self._calculate_sim_fluxes_derivative()
            flux_diff = self._calculate_difference_sim_exp_fluxes()
            flux_inv_cov = self.model.measured_fluxes_inv_cov
            grad2 = flux_der.T@flux_inv_cov@flux_diff
            
            return grad1 + grad2
        
        self.df = df2 if self.fit_measured_fluxes else df1
        
    
    def build_hessian(self):
        
        def _ddf(p):
            u, c = p[:self.nfreefluxes], p[self.nfreefluxes:]
            self.model.total_fluxes[:] = self.N@u
            self.model.concentrations[:] = c
            
            _, MDV_der = self._calculate_sim_MDVs_derivative()
            MDV_inv_cov = self.model.measured_inst_MDVs_inv_cov
            hess = MDV_der.T@MDV_inv_cov@MDV_der
            
            return hess
            
        def ddf1(p):
            return _ddf(p)
        
        def ddf2(p):
            hess1 = _ddf(p)
            
            flux_der = self._calculate_sim_fluxes_derivative()
            flux_inv_cov = self.model.measured_fluxes_inv_cov
            hess2 = flux_der.T@flux_inv_cov@flux_der
            
            return hess1 + hess2
        
        self.ddf = ddf2 if self.fit_measured_fluxes else ddf1
    
    
    def build_flux_and_conc_bound_constraints(self):
        
        A1 = self.N
        A2 = self.T@self.N
        A3 = -A2
        A4 = np.eye(self.nconcs)
        
        b1 = np.zeros(self.ntotalfluxes)
        vnet_lb, vnet_ub = np.array(list(self.model.net_fluxes_range.values())).T
        b2 = vnet_lb
        b3 = -vnet_ub
        b4 = np.zeros(self.nconcs)
        
        A = np.zeros(
            (self.ntotalfluxes+2*self.nnetfluxes+self.nconcs, 
             self.nfreefluxes+self.nconcs)
        )
        A[:(self.ntotalfluxes+2*self.nnetfluxes), :self.nfreefluxes] = np.vstack((A1, A2, A3))
        A[(self.ntotalfluxes+2*self.nnetfluxes):, self.nfreefluxes:] = A4
        b = np.concatenate((b1, b2, b3, b4))
        
        if self.solver == 'slsqp':
            self.constrs = {'type': 'ineq', 'fun': lambda p: A@p - b}
        elif self.solver == 'ralg':
            self.A = -A
            self.b = -b
    
    
    def build_initial_flux_and_conc_values(self, ini_netfluxes = None, ini_concs = None):
        '''
        Parameters
        ----------
        ini_netfluxes: array
            Initial guess of net fluxes.
        ini_concs: array
            Initial guess of concentrations.
        '''
        
        if ini_netfluxes is None:
            vnet_lb, vnet_ub = np.array(list(self.model.net_fluxes_range.values())).T
            vnet_ini = np.random.uniform(low = vnet_lb, high = vnet_ub)
        else:
            vnet_ini = ini_netfluxes
        u_ini = pinv2(self.T@self.N)@vnet_ini
        
        if ini_concs is None:
            c_lb, c_ub = np.array(list(self.model.concentrations_range.values())).T
            c_ini = np.random.uniform(low = c_lb, high = c_ub)
        else:
            c_ini = ini_concs

        self.x0 = np.concatenate((u_ini, c_ini))
    
    
    def _initialize_total_fluxes_and_concs(self):
        
        for fluxid in self.model.totalfluxids:
            self.model.total_fluxes[fluxid] = 0.0
    
        for metabid in self.model.concids:
            self.model.concentrations[metabid] = 0.0
            
    
    def _calculate_residuals(self):
        
        MDV_diff = self._calculate_difference_sim_exp_MDVs()
        MDV_diag = np.diag(self.model.measured_inst_MDVs_inv_cov)**0.5
        resids1 = MDV_diff*MDV_diag
        
        if self.fit_measured_fluxes:    
            flux_diff = self._calculate_difference_sim_exp_fluxes()
            flux_diag = np.diag(self.model.measured_fluxes_inv_cov)**0.5
            resids2 = flux_diff*flux_diag
            resids = np.concatenate((resids1, resids2))
        
        return resids
        

    def _get_exp_and_sim_inst_MDVs(self):

        exp_inst_MDVs = self.model.measured_inst_MDVs

        sim_inst_MDVs_all = self.calculator._calculate_inst_MDVs()
        self.calculator._build_initial_sim_MDVs()
        sim_inst_MDVs = {}
        for emuid in self.model.target_EMUs:
            instMDVs = self.model.initial_sim_MDVs[emuid].copy()   
            instMDVs.update({t: sim_inst_MDVs_all[emuid][t] 
                             for t in sim_inst_MDVs_all[emuid]})
            sim_inst_MDVs[emuid] = instMDVs

        return exp_inst_MDVs, sim_inst_MDVs 
        

    def solve_flux(self, tol = 1e-6, max_iters = 400, disp = False):
        
        self._initialize_total_fluxes_and_concs()

        if self.solver == 'slsqp':
            opt_obj, opt_p, is_success = self._solve_flux_slsqp(tol, max_iters, disp)
        elif self.solver == 'ralg':
            opt_obj, opt_p, is_success = self._solve_flux_ralg(tol, max_iters, disp)
        else:
            raise ValueError('currently only "slsqp" and "ralg" are acceptable')    

        opt_u = opt_p[:self.nfreefluxes]
        opt_c = opt_p[self.nfreefluxes:]
        
        opt_totalfluxes = pd.Series(self.N@opt_u, index = self.model.totalfluxids)
        opt_netfluxes = pd.Series(self.T@self.N@opt_u, index = self.model.netfluxids)        
        opt_concs = pd.Series(opt_c, index = self.model.concids)
        
        opt_resids = self._calculate_residuals()
        
        exp_inst_MDVs, sim_inst_MDVs = self._get_exp_and_sim_inst_MDVs()
        exp_fluxes, sim_fluxes = self._get_exp_and_sim_fluxes()
        
        nmeas = self._get_nmeasurements(opt_resids)
        nparams = self._get_nparameters(opt_p)
        
        hess = self._get_hessian(opt_p)

        _, dxsim_dp = self._calculate_sim_MDVs_derivative()
        dxsim_du = dxsim_dp[:, :self.nfreefluxes]
        
        dvsim_dp = self._calculate_sim_fluxes_derivative()
        dvsim_du = dvsim_dp[:, :self.nfreefluxes]
        
        return (
            opt_totalfluxes, 
            opt_netfluxes, 
            opt_concs, 
            opt_obj, 
            opt_resids, 
            nmeas, 
            nparams, 
            sim_inst_MDVs, 
            exp_inst_MDVs, 
            sim_fluxes, 
            exp_fluxes, 
            hess, 
            self.N, 
            self.T, 
            dxsim_du, 
            dvsim_du, 
            self.model.measured_inst_MDVs_inv_cov, 
            self.model.measured_fluxes_inv_cov, 
            is_success
        )
        


'''Define the MDV class.'''


__author__ = 'Chao Wu'


import re
from functools import lru_cache
from collections import OrderedDict, Counter
from collections.abc import Iterable
from itertools import combinations_with_replacement
from numbers import Real
import numpy as np
from scipy.special import comb
from scipy.linalg import inv
import warnings
warnings.filterwarnings('ignore', category = RuntimeWarning)


natAbuns = {'H': [0.999885, 0.000115],
            'C': [0.9893, 0.0107],
            'N': [0.99632, 0.00368],
            'O': [0.99757, 0.00038, 0.00205],
            'Si': [0.922297, 0.046832, 0.030872],
            'S': [0.9493, 0.0076, 0.0429, 0.0002]}
# ref. https://www.chem.ualberta.ca/~massspec/atomic_mass_abund.pdf


class MDV():
    '''
    Define MDV (i.e., mass isotopomer distribution vector) object and its operations.

    Convolution between mdv1 and mdv2 can be performed in three ways: 
    1. mdv1.conv(mdv2);
    2. mdv1*mdv2;
    3. mdv.conv(mdv1, mdv2).
    
    The zero element for convolution is MDV([0]), and the identity element is MDV([1]).
    
    Scalar multiplication (a*mdv) and MDV addition (mdv1 + mdv2) are also supported.
    In these cases, the resulting MDV are not automatically normalized.
    
    In addition to C, MDV can be built based on H, N, O, Si, and S.

    Parameters
    ----------
    fractions: list or array
        MDV vector.
    nonnegative: bool
        Whether to keep the elements >= 0.
    normalize: bool
        Whether to normalize MDV vector to ensure the sum equals one.
    base_atom: str
        Base atom for MDV.
            
    Attributes
    ----------
    value: array
        MDV vector.
    n_atoms: int
        # of atom.
    base_atom: str
        Base atom for MDV.
    fl (or fractional_labeling): float
        Fractional labeling.
    '''
    
    def __init__(
            self, 
            fractions, 
            nonnegative = True, 
            normalize = True, 
            base_atom = 'C'
    ):
        '''
        Parameters
        ----------
        fractions: list or array
            MDV vector.
        nonnegative: bool
            Whether to keep the elements >= 0.
        normalize: bool
            Whether to normalize MDV vector to ensure the sum == 1.
        base_atom: str
            Base atom for MDV.
        '''
        
        self.value = np.array(fractions)
        
        if nonnegative:
            self.value[self.value < 0] = 0
        
        if normalize and self.value.sum() != 0:
            self.value = self.value / self.value.sum()
            
        self.base_atom = base_atom
    

    def __iter__(self):
        
        return self.value
        
    
    def __getitem__(self, key):
        '''
        Parameters
        ----------
        key: int or slice
            Index.
        '''
        
        return self.value[key]
        
        
    def __array__(self):
        
        return self.value
    
    
    def conv(self, mdv):
        '''
        Note. Assume that the mdv convolved has the same base atom.

        Parameters
        ----------
        mdv: list or array or MDV 
        '''
        
        if not isinstance(mdv, self.__class__):
            mdv = self.__class__(mdv)
        
        mdvConv = gen_conv(self.value, mdv.value)
            
        return self.__class__(mdvConv, base_atom=self.base_atom)
    
        
    def __mul__(self, other):
        '''
        Parameters
        ----------
        other: scalar, list, array or MDV
        '''
        
        if isinstance(other, Real):
            return self.__class__(
                other*self.value, 
                nonnegative = False, 
                normalize = False, 
                base_atom=self.base_atom
            )                
        elif isinstance(other, Iterable):   
            return self.conv(other)    
        else:
            return NotImplemented
            
    
    def __rmul__(self, other):
        '''
        Parameters
        ----------
        other: MDV
        '''
        
        return self * other
        
        
    def __add__(self, other):
        '''
        Parameters
        ----------
        other: MDV
        '''
        
        if isinstance(other, type(self)):
            return self.__class__(
                self.value+other.value, 
                nonnegative = False, 
                normalize = False, 
                base_atom=self.base_atom
            )
        else:
            return NotImplemented
        
        
    def __radd__(self, other):
        '''
        Parameters
        ----------
        other: MDV
        '''
        
        return self + other
    
    
    def _correction_matrix(self, X, n_Xs):
        '''
        Parameters
        ----------
        X: str
            Which element the correction matrix will be generated for.
        n_Xs: int
            # of X atoms in metabolite or metabolite fragment, these atoms will be corrected.
        '''
        
        corrMat = np.zeros((self.n_atoms+1, self.n_atoms+1))
        
        natMDV = get_natural_MDV(n_Xs, base_atom = X) 
        
        for i in range(self.n_atoms+1):
            for j in range(i+1):    
                if i - j <= natMDV.n_atoms:
                    corrMat[i, j] = natMDV.value[i-j] 
                else:
                    corrMat[i, j] = 0
                    
        return corrMat
                
    
    def correct_for_natural_abundance(self, atom_dict):
        '''
        Parameters
        ----------
        atom_dict: dict
            element needs to be corrected => # of corresponding atoms in metabolite (fragment).
        '''
        
        corrMat = np.eye(self.n_atoms+1)
        for X, n_Xs in atom_dict.items():
            corrMatX = self._correction_matrix(X, n_Xs)
            corrMat = np.dot(corrMat, corrMatX)
        
        corrMDV = np.dot(inv(corrMat), self.value)

        return self.__class__(corrMDV, base_atom=self.base_atom)
        
        
    def correct_for_inoculum(self, fraction):
        '''
        Parameters
        ----------
        fraction: float in [0, 1]
            Fraction of inoculum in biomass measured.
        '''
        
        natMDV = get_natural_MDV(self.n_atoms, base_atom = self.base_atom)
        corrMDV =  (self.value - fraction*natMDV.value)/(1 - fraction)

        return self.__class__(corrMDV, base_atom=self.base_atom)
    
    
    @property
    @lru_cache()
    def n_atoms(self):
        
        return self.value.size - 1
        
        
    @property
    @lru_cache()
    def fl(self):
    
        if self.n_atoms == 0:
            raise ValueError('no atom in MDV')
        else:
            return round((self.value*np.arange(self.n_atoms+1)).sum()/self.n_atoms, 4)


    fractional_labeling = fl
        
        
    def __repr__(self):
    
        return f'{self.__class__.__name__}([{", ".join(map(str, self.value.round(3)))}])'

    
def _isotopomer_combination(n_atoms, n_natural_isotops):
    '''
    Parameters
    ----------    
    n_atoms: int
        # of atoms.
    n_natural_isotops: int
        # of natural isotopmers.
    
    Returns
    -------
    combs2: dict
    
    Example
    -------
    >>> _isotopomer_combination(2, 3)
    OrderedDict([(0, [Counter({0: 2})]), (1, [Counter({0: 1, 1: 1})]), (2, [Counter({0: 1, 2: 1}), 
    Counter({1: 2})]), (3, [Counter({1: 1, 2: 1})]), (4, [Counter({2: 2})])])
    '''
    
    allCombos = combinations_with_replacement(range(n_natural_isotops), n_atoms)   
    
    combos1 = {}
    for combo in allCombos:
        combos1.setdefault(sum(combo), []).append(combo)
    
    combos2 = OrderedDict()
    for mass in sorted(combos1):
        for combo in combos1[mass]:        
            combos2.setdefault(mass, []).append(Counter(combo))
    
    return combos2
    
    
def get_natural_MDV(n_atoms, base_atom = 'C'):
    '''
    Calculate the MDV of an unlabeled fragment.

    Parameters
    ----------
    n_atoms: int
        # of atoms.
    base_atom: str
        Base atom for MDV.
    '''
    
    natAbun = natAbuns[base_atom]
    
    combos = _isotopomer_combination(n_atoms, len(natAbun))
    
    mdv = []
    for _, counters in combos.items():    
        ele = 0
        for counter in counters:
            item = 1
            left = n_atoms
            for isotop, count in counter.items():
                item *= comb(left, count) * natAbun[isotop]**count
                left -= count
            ele += item
        mdv.append(ele)
    
    return MDV(mdv, base_atom=base_atom)


def get_substrate_MDV(atom_nos, labeling_pattern, percentage, purity, label_atom='C'):
    '''
    Calculate the MDV of a fragment from labeled substrate. 
    Currently this function only supports H, C or N-labeled substrate.
    
    Parameters
    ----------
    atom_nos: list of int
        Atom NOs, starting from 1. 
    labeling_pattern: str or list of str
        Labeling pattern of substrate, '0' for unlabeled atom, '1' for labeled atom, 
        e.g., '100000' for 1-13C glucose. 
        
        List if tracer with multiple labeling patterns are used. 
        
        Natural substrate (with all '0's) don't need to be explicitly set.
        
        If str, labeling_pattern should not be natural substrate.
    percentage: float or list of float
        Molar percentage (in range of [0,1]) of corresponding tracer. 
        Sum of percentage should be <= 1, and the rest will be considered as natural substrate.
        
        List if tracer with multiple labeling patterns are used. 
        
        * If list, len(percentage) should be equal to len(labeling_pattern).
        * If float, labeling_pattern should not be natural substrate.
    purity: float or list of float
        Labeled atom purity (in range of [0,1]) of corresponding tracer.
            
        List if tracer with multiple labeling patterns are used.

        * If list, len(purity) should be equal to len(labeling_pattern).
        * If float, labeling_pattern should not be natural substrate.
    label_atom: str
        Labeled atom, i.e., base atom in the MDV. Currently supports only "H", "C" 
        and "N".
    '''
    
    if not isinstance(labeling_pattern, list):
        
        if re.match(r'^0+$', labeling_pattern):
            raise ValueError('use mdv.get_natural_MDV for natural substrate')

        labeling_pattern = [labeling_pattern]
        
    if not isinstance(percentage, list):
        percentage = [percentage]
        
    if sum(percentage) > 1:
        raise ValueError('sum of tracer percentage should be <= 1')

    if not isinstance(purity, list):
        purity = [purity]
    
    nAtoms = len(atom_nos)
    
    singleAtomMdv = {}
    for pat, pur, in zip(labeling_pattern, purity):
        singleAtomMdv[pat] = {
            '1': MDV([1-pur, pur], base_atom=label_atom),
            '0': get_natural_MDV(1, base_atom = label_atom)
        }
    
    
    mdvAllTracer = MDV(np.zeros(nAtoms+1), base_atom=label_atom)
    for pat, per, pur in zip(labeling_pattern, percentage, purity):
        
        mdvPerTracer = MDV([1], base_atom=label_atom)
        for atomNO in atom_nos:
            mdvPerTracer *= singleAtomMdv[pat][pat[atomNO-1]]
        
        mdvAllTracer += per*mdvPerTracer     
    
    mdvNatural = (1 - sum(percentage))*get_natural_MDV(nAtoms, base_atom = label_atom)

    return mdvAllTracer + mdvNatural
    
    
def gen_conv(arr1, arr2):
    '''
    Parameters
    ----------
    arr1, arr2: list or array
        If one of the arrs is 2-D, the function performs 2-D convolution of 
        MDV (array) and MDV derivative.
    '''
    
    nAtoms1 = len(arr1) - 1
    nAtoms2 = len(arr2) - 1
    
    if nAtoms2 > nAtoms1:
        arr1, arr2 = arr2, arr1
        nAtoms1, nAtoms2 = nAtoms2, nAtoms1
    
    arrConv = []
    for i in range(nAtoms1+nAtoms2+1):
        if i <= nAtoms2:
            arrConv.append(sum([arr1[i-j] * arr2[j] for j in range(i+1)]))    
        elif nAtoms2 < i <= nAtoms1:
            arrConv.append(sum([arr1[i-j] * arr2[j] for j in range(nAtoms2+1)]))
        else:
            arrConv.append(sum([arr1[i-j] * arr2[j] for j in range(i-nAtoms1, nAtoms2+1)]))
        
    return np.array(arrConv)    
    
    
def conv(mdv1, mdv2):
    '''
    Perform convolution between two MDVs.

    Parameters
    ----------
    mdv1: list or array or MDV
    mdv2: list or array or MDV
    
    Returns
    -------
    mdv: MDV
    '''
    
    if not isinstance(mdv1, MDV):
        mdv1 = MDV(mdv1)
    
    if not isinstance(mdv2, MDV):
        mdv2 = MDV(mdv2)
    
    return mdv1 * mdv2
        

def diff_conv(mdv_mdvder1, mdv_mdvder2):
    '''
    Parameters
    ----------
    mdv_mdvder1: 2-list of arrays
        namely [MDV, MDVder], MDVder in shape of (len(MDV), len(v))
    mdv_mdvder2:: 2-list of arrays
        namely [MDV, MDVder], MDVder in shape of (len(MDV), len(v))
        
    Returns
    -------
    mdv: array
    mdvder: 2-D array
        In shape of (len(MDV), len(v)).
    '''
    
    mdv1, mdvder1 = mdv_mdvder1
    mdv2, mdvder2 = mdv_mdvder2
    
    if isinstance(mdv1, MDV):
        mdv1 = mdv1.value
        
    if isinstance(mdv2, MDV):
        mdv2 = mdv2.value
        
    mdv = gen_conv(mdv1, mdv2)
    mdvder = gen_conv(mdv1, mdvder2) + gen_conv(mdv2, mdvder1)
    
    return mdv, mdvder



'''Define the Fitter class.'''


__author__ = 'Chao Wu'


from collections.abc import Iterable
from functools import partial
from copy import deepcopy
from math import ceil
import numpy as np
from multiprocessing import Pool
from ..optim.optim import Optimizer
from .simulate import Simulator
from ..io.inputs import read_measurements_from_file, read_initial_values
from ..io.results import FitResults, FitMCResults
from ..utils.utils import Calculator
from ..utils.progress import Progress
from ..solver.nlpsolver import MFAModel 


class Fitter(Optimizer, Simulator):
    '''
    Parameters
    ----------
    model: Model
        Freeflux Model.
    '''
    
    def __init__(self, model):
        '''
        Parameters
        ----------
        model: Model
            Freeflux Model.
        '''
        
        super().__init__(model)
        
        self.calculator = Calculator(self.model)
        
    
    def set_measured_MDV(self, fragmentid, mean, sd):
        '''
        Set measured MDV.
        
        Parameters
        ----------
        fragmentid: str
            Metabolite ID + "_" + atom NOs, e.g., "Glu_12345".
        mean: array
            Means of measured MDV vector.
        sd: array
            Standard deviations of measured MDV vector.
        '''
        
        self.model.measured_MDVs[fragmentid] = [np.array(mean), np.array(sd)]
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_measured_MDVs, fragmentid))
            
            
    def set_measured_MDVs_from_file(self, file):
        '''
        Read measured MDVs from file.
        
        Parameters
        ----------
        file: file path
            tsv or excel file with fields "fragment_ID", "mean" and "sd". 
            "fragment_ID" is metabolite ID + "_" + atom NOs, e.g., 'Glu_12345'; 
            "mean" and "sd" are the mean and standard deviation of MDV with 
            element seperated by ",".

            Header line starts with "#", and will be skiped.
        '''
        
        measMDVs = read_measurements_from_file(file)

        for emuid, [mean, sd] in measMDVs.iterrows():
            self.model.measured_MDVs[emuid] = [
                np.array(list(map(float, mean.split(',')))),
                np.array(list(map(float, sd.split(','))))
            ]
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(
                partial(self._unset_measured_MDVs, measMDVs.index.tolist())
            )
        
        
    def _unset_measured_MDVs(self, fragmentids):
        '''
        Parameters
        ----------
        fragmentids: str or list of str
            Measured MDV ID(s).
        '''
        
        if not isinstance(fragmentids, Iterable):
            fragmentids = [fragmentids]
            
        for fragmentid in fragmentids:
            if fragmentid in self.model.measured_MDVs:
                self.model.measured_MDVs.pop(fragmentid)
        
        
    def set_measured_flux(self, fluxid, mean, sd):
        '''
        Set measured flux.
        
        Parameters
        ----------
        fluxid: str
            Flux ID, i.e., reaction ID. Typically, measured fluxes are substrate 
            consumption product formation or cell growth, etc. They should be irreversible.
        mean: float
            Mean of measured flux.
        sd: float
            Standard deviation of measured flux.
        '''
        
        self.model.measured_fluxes[fluxid] = [mean, sd]
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_measured_fluxes, fluxid))
            
            
    def set_measured_fluxes_from_file(self, file):
        '''
        Read measured fluxes from file.
        
        Parameters
        ----------
        file: file path
            tsv or excel file with fields "reaction_ID", "mean" and "sd".
            "reaction_ID" is reaction ID, typically measured fluxes are substrate 
            consumption, product formation or cell growth, etc. They should be irreversible;
            "mean" and "sd" are the mean and standard deviation of measured flux.

            Header line starts with "#", and will be skiped.
        '''
        
        measFluxes = read_measurements_from_file(file)
            
        for emuid, [mean, sd] in measFluxes.iterrows():    
            self.model.measured_fluxes[emuid] = [mean, sd]
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(
                partial(self._unset_measured_fluxes, measFluxes.index.tolist())
            )
    
    
    def _unset_measured_fluxes(self, fluxids):
        '''
        Parameters
        ----------
        fluxids: str or list of str
            Measured flux ID(s).
        '''
        
        if not isinstance(fluxids, Iterable):
            fluxids = [fluxids]
            
        for fluxid in fluxids: 
            if fluxid in self.model.measured_fluxes:
                self.model.measured_fluxes.pop(fluxid)
            
        
    def set_unbalanced_metabolites(self, metabids):
        '''
        Parameters
        ----------
        metabids: str or list of str
            Unbalanced metabolite ID(s).
        '''
        
        if not isinstance(metabids, Iterable):
            metabids = [metabids]
        
        self.model.unbalanced_metabolites.update(metabids)
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(
                partial(self._unset_unbalanced_metabolites, metabids)
            )
        
        
    def _unset_unbalanced_metabolites(self, metabids):
        '''
        Parameters
        ----------
        metabids: list of str
            Unbalanced metabolite IDs.
        '''
        
        for metabid in metabids:
            if metabid in self.model.unbalanced_metabolites:
                self.model.unbalanced_metabolites.remove(metabid)
            
    
    def set_flux_bounds(self, fluxid, bounds):
        '''
        Set lower and upper bounds of flux.
        
        Parameters
        ----------
        fluxid: str or 'all'
            Flux ID, i.e., reaction ID. Since forward and backward fluxes of 
            reversible reaction are usually unknown, the method is used to set 
            the range of net fluxes.
            
            If 'all', all fluxes will be set to the range.
        bounds: 2-list
            [lower bound, upper bound]. Lower bound is not allow to equal upper bound.
            Use set_measured_flux (or set_measured_fluxes_from_file) to set fixed value
            of flux.

            For irreversible reaction, the lower bound will be set to zero ignorant 
            of bounds[0].
        '''
        
        fluxids = []
        if bounds[0] < bounds[1]:   # lower bound not allow to equal upper bound
            if fluxid == 'all':
                for rxnid in self.model.reactions:
                    self._set_flux_bounds(rxnid, bounds)
                    fluxids.append(rxnid)    
            elif fluxid in self.model.reactions:
                self._set_flux_bounds(fluxid, bounds)
                fluxids = [fluxid]
            else:
                raise ValueError(f'flux range set to nonexistent reaction {fluxid}')
        else:
            raise ValueError(
                'lower bound of flux should be less than upper bound, '
                'use set_measured_flux or set_measured_fluxes_from_file '
                'to set fixed flux value'
            )
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_flux_bounds, fluxids))
    
    
    def _decompose_network(self, n_jobs):
        '''
        Parameters
        ----------
        n_jobs: int
            # of jobs to run in parallel.
        '''
        
        if not self.model.measured_MDVs:
            raise ValueError('call set_measured_MDV or set_measured_MDVs_from_file first')
        
        if not self.model.EAMs:
            if n_jobs <= 0:
                raise ValueError('n_jobs should be a positive value')    
            else:
                self.model.target_EMUs = list(self.model.measured_MDVs.keys())
                
                metabids = []
                atom_nos = []
                for emuid in self.model.target_EMUs:
                    metabid, atomNOs = emuid.split('_')
                    metabids.append(metabid)
                    atom_nos.append(atomNOs)
                
                EAMs = self.model._decompose_network(metabids, atom_nos, n_jobs = n_jobs)
                for size, EAM in EAMs.items():
                    self.model.EAMs[size] = EAM
                
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_decomposition)
            
            
    def _unset_decomposition(self):
        
        self.model.EAMs.clear()        
    
    
    def _calculate_matrix_As_and_Bs_derivatives_p(self, kind, n_jobs):
        '''
        Parameters
        ----------
        kind: {"ss", "inst"}
            * If "ss", variables are free fluxes only.
            * If "inst", variables include both free fluxes and concentrations.
        '''
        
        if not self.model.matrix_As_der_p or not self.model.matrix_Bs_der_p:
            self.calculator._calculate_matrix_As_and_Bs_derivatives_p(kind, n_jobs)
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_matrix_As_and_Bs_derivatives_p)
        
        
    def _unset_matrix_As_and_Bs_derivatives_p(self):
        
        self.model.matrix_As_der_p.clear()
        self.model.matrix_Bs_der_p.clear()
    
    
    def _calculate_substrate_MDV_derivatives_p(self, kind, extra_subs = None):
        '''
        Parameters
        ----------
        kind: {"ss", "inst"}
            * If "ss", variables are free fluxes only.
            * If "inst", variables include free fluxes and concentrations.
        extra_subs: str or list of str
            Metabolite ID(s), additional metabolites considered as substrates.    
        '''
        
        if extra_subs is not None and not isinstance(extra_subs, list):
            extra_subs = [extra_subs]

        if not self.model.substrate_MDVs_der_p:
            self.calculator._calculate_substrate_MDV_derivatives_p(kind, extra_subs)
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_substrate_MDV_derivatives_p)
            
    
    def _unset_substrate_MDV_derivatives_p(self):
        
        self.model.substrate_MDVs_der_p.clear()
        
    
    def _calculate_null_space(self):
        
        if self.model.null_space is None:
            self.calculator._calculate_null_space()
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_null_space)
            
            
    def _unset_null_space(self):
        
        self.model.null_space = None
            
    
    def _calculate_transform_matrix(self):
        
        if self.model.transform_matrix is None:
            self.calculator._calculate_transform_matrix()
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_transform_matrix)    
    
            
    def _unset_transform_matrix(self):
        
        self.model.transform_matrix = None
    
    
    def _calculate_measured_MDVs_inversed_covariance_matrix(self):
    
        if self.model.measured_MDVs_inv_cov is None:
            self.calculator._calculate_measured_MDVs_inversed_covariance_matrix()
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_measured_MDVs_inversed_covariance_matrix)    
    
    
    def _unset_measured_MDVs_inversed_covariance_matrix(self):
        
        self.model.measured_MDVs_inv_cov = None
        
    
    def _calculate_measured_fluxes_inversed_covariance_matrix(self):
        
        if self.model.measured_fluxes_inv_cov is None:
            self.calculator._calculate_measured_fluxes_inversed_covariance_matrix()
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_measured_fluxes_inversed_covariance_matrix)
    
            
    def _unset_measured_fluxes_inversed_covariance_matrix(self):
        
        self.model.measured_fluxes_inv_cov = None
        
    
    def _calculate_measured_fluxes_derivative_p(self, kind):
        '''
        Parameters
        ----------
        kind: {"ss", "inst"}
            * If "ss", variables are free fluxes only.
            * If "inst", variables include free fluxes and concentrations.
        '''
        
        if self.model.measured_fluxes_der_p is None:
            self.calculator._calculate_measured_fluxes_derivative_p(kind)
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_measured_fluxes_derivative_p)
        
            
    def _unset_measured_fluxes_derivative_p(self):
        
        self.model.measured_fluxes_der_p = None
    
    
    def _estimate_fluxes_range(self, exclude_metabs = None):
        '''
        Parameters
        ----------
        exclude_metabs: list
            Metabolite IDs, metabolites excluded from mass balance.
        '''
        
        fluxids = []
        if not self.model.net_fluxes_range:
            FVAres = self.estimate_fluxes_range(
                exclude_metabs = exclude_metabs, 
                show_progress = False
            )
            for fluxid, fluxRange in FVAres.flux_ranges.items():
                self.model.net_fluxes_range[fluxid] = fluxRange
                fluxids.append(fluxid)
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_net_fluxes_range, fluxids))
            
            
    def _unset_net_fluxes_range(self, fluxids):
        '''
        Parameters
        ----------
        fluxids: str or list of str
            Fluxe ID(s).
        '''
        
        if not isinstance(fluxids, Iterable):
            fluxids = [fluxids]
        
        for fluxid in fluxids:
            if fluxid in self.model.net_fluxes_range:
                self.model.net_fluxes_range.pop(fluxid)
        
    
    def prepare(self, dilution_from = None, n_jobs = 1):
        '''
        Parameters
        ----------
        dilution_from: str or list of str
            ID(s) of unlabeled (inactive) metabolite leading to dilution effect.
            These metabolites have zero stoichiometric coefficients in reaction network.
        n_jobs: int
            If n_jobs > 1, preparation will run in parallel.
        '''
        
        self._decompose_network(n_jobs)
        self._calculate_null_space()
        self._calculate_transform_matrix()
        self._lambdify_matrix_As_and_Bs()
        self._calculate_matrix_As_and_Bs_derivatives_p('ss', n_jobs)
        self._calculate_substrate_MDVs(dilution_from)
        self._calculate_substrate_MDV_derivatives_p('ss', dilution_from)
        self._calculate_measured_MDVs_inversed_covariance_matrix()
        self._calculate_measured_fluxes_inversed_covariance_matrix()
        self._calculate_measured_fluxes_derivative_p('ss')
        self._set_default_flux_bounds()
        self._estimate_fluxes_range(self.model.unbalanced_metabolites)
        
    
    def _check_dependencies(self, fit_measured_fluxes):
        '''
        Parameters
        ----------
        fit_measured_fluxes: bool
            Whether to fit measured fluxes.
        '''

        if not self.model.net_fluxes_bounds:
            raise ValueError('call set_flux_bounds first')
        if not self.model.measured_MDVs:
            raise ValueError('call set_measured_MDV or set_measured_MDVs_from_file first')
        if not self.model.measured_fluxes:
            raise ValueError('call set_measured_flux or set_measured_fluxes_from_file first')
        if not self.model.labeling_strategy:
            raise ValueError('call labeling_strategy first')
        
        checklist = [
            not self.model.target_EMUs, 
            self.model.transform_matrix is None, 
            self.model.null_space is None,
            self.model.measured_MDVs_inv_cov is None, 
            not self.model.matrix_As, 
            not self.model.matrix_Bs,
            not self.model.matrix_As_der_p, 
            not self.model.matrix_Bs_der_p, 
            not self.model.substrate_MDVs,
            not self.model.substrate_MDVs_der_p, 
            self.model.measured_fluxes_der_p is None
        ]
        if fit_measured_fluxes:
            checklist.append(self.model.measured_fluxes_inv_cov is None)
        
        if any(checklist):
            raise ValueError('call prepare first')


    def solve(
            self, 
            fit_measured_fluxes = True, 
            ini_fluxes = None, 
            solver = 'slsqp', 
            tol = 1e-6, 
            max_iters = 400, 
            show_progress = True
    ):
        '''
        Parameters
        ----------
        fit_measured_fluxes: bool
            Whether to fit measured fluxes.
        ini_fluxes: ser or file in .tsv or .xlsx
            Initial values of net fluxes
        solver: {"slsqp", "ralg"}
            * If "slsqp", scipy.optimize.minimze will be used.
            * If "ralg", openopt NLP solver will be used.
        tol: float
            Tolerance for termination.
        max_iters: int
            Maximum # of iterations.
        show_progress: bool
            Whether to show the progress bar.    
        '''
        
        self._check_dependencies(fit_measured_fluxes)

        if ini_fluxes is not None:
            iniFluxes = read_initial_values(ini_fluxes, self.model.netfluxids)
        else:
            iniFluxes = ini_fluxes
            
        optModel = MFAModel(self.model, fit_measured_fluxes, solver)
        optModel.build_objective()
        optModel.build_gradient()
        optModel.build_flux_bound_constraints()
        optModel.build_initial_flux_values(ini_netfluxes = iniFluxes)
        
        with Progress('fitting', silent = not show_progress):
            res = optModel.solve_flux(tol, max_iters)

        return FitResults(
            *res[:7], 
            deepcopy(res[7]), 
            res[8], 
            deepcopy(res[9]), 
            *res[10:])   
    
    
    def _solve_with_confidence_intervals(
            self, 
            fit_measured_fluxes, 
            ini_fluxes, 
            solver, 
            tol, 
            max_iters, 
            nruns
    ):
        '''
        Parameters
        ----------
        fit_measured_fluxes: bool
            Whether to fit measured fluxes.
        ini_fluxes: ser or file in .tsv or .xlsx or None
            Initial values of net fluxes.   
        solver: {"slsqp", "ralg"}
            * If "slsqp", scipy.optimize.minimze will be used.
            * If "ralg", openopt NLP solver will be used.
        tol: float
            Tolerance for termination.
        max_iters: int
            Maximum # of iterations.
        nruns: int
            # of estimations in each worker. 
        '''
        
        import platform
        if platform.system() == 'Linux':
            import os
            os.sched_setaffinity(os.getpid(), range(os.cpu_count()))

        self._lambdify_matrix_As_and_Bs()   
        
        if ini_fluxes is not None:
            iniFluxes = read_initial_values(ini_fluxes, self.model.netfluxids)
        else:
            iniFluxes = ini_fluxes

        optTotalfluxesSet = []
        optNetfluxesSet = []
        for _ in range(nruns):
            self.calculator._generate_random_fluxes()
            self.calculator._generate_random_MDVs()
            
            optModel = MFAModel(self.model, fit_measured_fluxes, solver)
            optModel.build_objective()
            optModel.build_gradient()
            optModel.build_flux_bound_constraints()
            optModel.build_initial_flux_values(ini_netfluxes = iniFluxes)

            try:
                while True:
                    (optTotalfluxes, 
                     optNetfluxes, 
                     *_, 
                     isSuccess
                    ) = optModel.solve_flux(tol, max_iters)
                    if isSuccess:
                        break
            except:
                continue
            optTotalfluxesSet.append(optTotalfluxes)
            optNetfluxesSet.append(optNetfluxes)
            
            self.calculator._reset_measured_fluxes()
            self.calculator._reset_measured_MDVs()
               
        return optTotalfluxesSet, optNetfluxesSet
        
    
    def solve_with_confidence_intervals(
            self, 
            fit_measured_fluxes = True, 
            ini_fluxes = None, 
            solver = 'slsqp', 
            tol = 1e-6, 
            max_iters = 400, 
            n_runs = 100, 
            n_jobs = 1, 
            show_progress = True
    ):
        '''
        Parameters
        ----------
        fit_measured_fluxes: bool
            Whether to fit measured fluxes.
        ini_fluxes: ser or file in .tsv or .xlsx
            Initial values of net fluxes.  
        solver: {"slsqp", "ralg"}
            * If "slsqp", scipy.optimize.minimze will be used.
            * If "ralg", openopt NLP solver will be used.
        tol: float
            Tolerance for termination.
        max_iters: int
            Maximum # of iterations.
        show_progress: bool
            Whether to show the progress bar.
        n_runs: int
            # of runs to estimate confidence intervals.
        n_jobs: int
            # of jobs to run in parallel.
        '''
        
        self._check_dependencies(fit_measured_fluxes)
        
        self._unset_matrix_As_and_Bs()

        if n_runs <= n_jobs:
            nruns_worker = 1
        else:
            nruns_worker = ceil(n_runs/n_jobs)
        
        pool = Pool(processes = n_jobs)
        with Progress('fitting with CIs', silent = not show_progress):
            fluxesSet = []
            for _ in range(n_jobs):
                fluxes = pool.apply_async(
                    func = self._solve_with_confidence_intervals, 
                    args = (
                        fit_measured_fluxes,
                        ini_fluxes,
                        solver,
                        tol,
                        max_iters,
                        nruns_worker
                    )
                )
                fluxesSet.append(fluxes)
            
            pool.close()    
            pool.join()
        
        fluxesSet = [fluxes.get() for fluxes in fluxesSet]
        
        totalFluxesSet = []
        netFluxesSet = []
        for totalFluxesSubset, netFluxesSubset in fluxesSet:
            totalFluxesSet.extend(totalFluxesSubset)
            netFluxesSet.extend(netFluxesSubset)

        self._lambdify_matrix_As_and_Bs()    
        
        return FitMCResults(totalFluxesSet, netFluxesSet)



'''Define classes of simulation and fitting results.'''


__author__ = 'Chao Wu'


import numpy as np
import pandas as pd
from scipy.linalg import pinv2
from ..core.mdv import MDV
from ..analysis.stats import (_chi2_test, _normal_probability, _simulated_vs_measured_MDVs, 
                              _simulated_vs_measured_fluxes, _simulated_vs_measured_inst_MDVs,
                              _confidence_intervals_le, _confidence_intervals_mc, _MDV_kinetics,
                              _contribution_matrix, _sensitivity)
import warnings
warnings.filterwarnings('ignore', category = RuntimeWarning)


class pDict(dict):

    def __init__(self, *args, digits = 3, **kwargs):
        
        super().__init__(*args, **kwargs)
        self.digits = digits
    
        
    def __repr__(self):
        
        itemStrs = []
        for k, v in self.items():
            if isinstance(v, MDV):
                itemStr = '%s: %s' % (k, v)
            elif isinstance(v, list):
                itemStr = '%s: [%s]' % (k, ', '.join(map(str, np.round(v, self.digits))))
            else:
                itemStr = '%s: %s' % (k, np.round(v, self.digits))
            itemStrs.append(itemStr)
        
        return '\n'.join(itemStrs)

    
    
    
class FBAResults():
    '''
    Parameters
    ----------
    obj: dict
        Reaction ID => coefficient, i.e., the objective function.
    opt_obj: float
        Optimal objective.
    opt_fluxes: OrderedDict
        Optimal fluxes.
    '''
        
    def __init__(self, obj, opt_obj, opt_fluxes):
        '''
        Parameters
        ----------
        obj: dict
            Reaction ID => coefficient, i.e., the objective function.
        opt_obj: float
            Optimal objective.
        opt_fluxes: OrderedDict
            Optimal fluxes.
        '''
        
        self._obj = obj
        self._opt_obj = opt_obj
        self._opt_fluxes = opt_fluxes
        
    
    @property
    def objective(self):
        
        objStr = ''
        count = 0
        for fluxid, coe in self._obj.items():
            if coe > 0 and count > 0:
                objStr += '+%s*%s' % (coe, fluxid)
            else:
                objStr += '%s*%s' % (coe, fluxid)
            count += 1    
                
        return objStr
    
    
    @property
    def opt_objective(self):
        
        return round(self._opt_obj, 3)
        
        
    @property
    def opt_fluxes(self):
        
        return pDict(self._opt_fluxes)
        
    
    def __repr__(self):
    
        return (
            f'objective: {self.objective}\n'
            f'optimal objective: {self.opt_objective}\n'
            f'optimal fluxes\n{self.opt_fluxes}'
        )
        



class FVAResults():
    '''
    Parameters
    ----------
    flux_ranges: dict
        Reaction ID => [lower bound, upper bound].
    '''
    
    def __init__(self, flux_ranges):
        '''
        Parameters
        ----------
        flux_ranges: dict
            Reaction ID => [lower bound, upper bound].
        '''
        
        self._flux_ranges = flux_ranges
    
        
    @property    
    def flux_ranges(self):
        
        return pDict(self._flux_ranges)
        
    
    def __repr__(self):
        
        return 'flux ranges\n%s' % self.flux_ranges
        
        
        
        
class SimResults():
    '''
    Parameters
    ----------
    simulated_MDVs: dict
        EMU ID => MDV.
    '''
    
    def __init__(self, simulated_MDVs):
        '''
        Parameters
        ----------
        simulated_MDVs: dict
            EMU ID => MDV.
        '''
        
        self._simulated_MDVs = simulated_MDVs
        self._simulated_EMUs = sorted(self._simulated_MDVs.keys())
    
    
    @property
    def simulated_EMUs(self):
        
        return self._simulated_EMUs
        
    
    def simulated_MDV(self, emuid):
        '''
        Parameters
        emuid: str
            EMU ID
        '''
        
        return self._simulated_MDVs[emuid]
        
        
    def __repr__(self):
        
        return 'simulated MDVs\n%s' % pDict(self._simulated_MDVs)
        
        
        
        
class InstSimResults():
    '''
    Parameters
    ----------
    simulated_inst_MDVs: dict
        EMU IDs => {timepoints => MDV}.
    '''
    
    def __init__(self, simulated_inst_MDVs):
        '''
        Parameters
        ----------
        simulated_inst_MDVs: dict
            EMU IDs => {timepoints => MDV}.
        '''
        
        self._simulated_inst_MDVs = simulated_inst_MDVs
        self._simulated_EMUs = sorted(self._simulated_inst_MDVs)
        self._timepoints = sorted(self._simulated_inst_MDVs[self._simulated_EMUs[0]])
        
    
    @property
    def simulated_EMUs(self):
        
        return self._simulated_EMUs
        
        
    @property
    def timepoints(self):
    
        return self._timepoints
        
    
    def simulated_MDV(self, emuid):
        '''
        Parameters
        ----------
        emuid: str
            EMU ID.
        '''
        
        return pDict(self._simulated_inst_MDVs[emuid])        
        
        
    def plot_MDV_kinetics(self, emuid, show_fig = True, output_dir = None):
        '''
        Parameters
        ----------
        emuid: str
            EMU ID.
        show_fig: bool
            Whether to show figure.
        output_dir: str
            Output directory.
        '''    
        
        _MDV_kinetics(emuid, self._simulated_inst_MDVs[emuid], show_fig, output_dir)
        
        
    def __repr__(self):
        
        mdvstrs = []
        for emuid in self.simulated_EMUs:
            mdvstr = '%s\n%s' % (emuid, self.simulated_MDV(emuid))
            mdvstrs.append(mdvstr)
            
        return 'simulated_MDVs\n' + '\n\n'.join(mdvstrs)
        
        
        
        
class FitResults():
    '''
    Parameters
    ----------
    opt_total_fluxes: ser
        Total fluxes at optimal objective.
    opt_net_fluxes: ser
        Net fluxes at optimal objective.
    opt_obj: float
        Optimal value of objective.
    opt_resids: array
        Optimal weighted residuals.
    n_meas: int
        # of measurements.
    n_params: int
        # of parameters.
    sim_MDVs: dict
        EMU ID => simulated MDV.
    exp_MDVs: dict
        EMU ID => [means, sds].
    sim_fluxes: dict
        Flux ID => simulated flux.
    exp_fluxes: dict
        Flux ID => [mean, sd].
    hessian: array
        Hessian matrix at convergence.
    null_space: array
        Null space of stoichiometric matrix.
    transform_matrix: array
        Transform matrix from total fluxes to net fluxes.
    sim_MDVs_der_u: array
        Derivative of simulated MDVs w.r.t. free fluxes.
    sim_fluxes_der_u: array
        Derivative of simualted fluxes w.r.t. free fluxes.
    exp_MDVs_inv_cov: array
        Inversed covariance matrix of measured MDVs.
    exp_fluxes_inv_cov: array
        Inversed covariance matrix of measured fluxes.
    is_success: bool
        Whether the optimization is successful.
    '''
    
    def __init__(
            self, 
            opt_total_fluxes, 
            opt_net_fluxes, 
            opt_obj, 
            opt_resids, 
            n_meas, 
            n_params, 
            sim_MDVs, 
            exp_MDVs, 
            sim_fluxes, 
            exp_fluxes, 
            hessian, 
            null_space, 
            transform_matrix, 
            sim_MDVs_der_u, 
            sim_fluxes_der_u, 
            exp_MDVs_inv_cov, 
            exp_fluxes_inv_cov, 
            is_success
    ):
        '''
        Parameters
        ----------
        opt_total_fluxes: ser
            Total fluxes at optimal objective.
        opt_net_fluxes: ser
            Net fluxes at optimal objective.
        opt_obj: float
            Optimal value of objective.
        opt_resids: array
            Optimal weighted residuals.
        n_meas: int
            # of measurements.
        n_params: int
            # of parameters.
        sim_MDVs: dict
            EMU ID => simulated MDV.
        exp_MDVs: dict
            EMU ID => [means, sds].
        sim_fluxes: dict
            Flux ID => simulated flux.
        exp_fluxes: dict
            Flux ID => [mean, sd].
        hessian: array
            Hessian matrix at convergence.
        null_space: array
            Null space of stoichiometric matrix.
        transform_matrix: array
            Transform matrix from total fluxes to net fluxes.
        sim_MDVs_der_u: array
            Derivative of simulated MDVs w.r.t. free fluxes.
        sim_fluxes_der_u: array
            Derivative of simualted fluxes w.r.t. free fluxes.
        exp_MDVs_inv_cov: array
            Inversed covariance matrix of measured MDVs.
        exp_fluxes_inv_cov: array
            Inversed covariance matrix of measured fluxes.
        is_success: bool
            Whether the optimization is successful.
        '''
        
        self._opt_total_fluxes = opt_total_fluxes
        self._opt_net_fluxes = opt_net_fluxes
        self._opt_obj = opt_obj
        self.opt_resids = opt_resids
        self.simulated_MDVs = sim_MDVs
        self.measured_MDVs = exp_MDVs
        self.simulated_fluxes = sim_fluxes
        self.measured_fluxes = exp_fluxes
        self.hessian = hessian
        self.null_space = null_space
        self.transform_matrix = transform_matrix
        self.n_meas = n_meas
        self.n_params = n_params
        self.dof = self.n_meas - self.n_params
        self.sim_MDVs_der_u = sim_MDVs_der_u
        self.sim_fluxes_der_u = sim_fluxes_der_u
        self.exp_MDVs_inv_cov = exp_MDVs_inv_cov
        self.exp_fluxes_inv_cov = exp_fluxes_inv_cov
        self.is_success = is_success
        
    
    @property
    def n_total_fluxes(self):
        
        return len(self._opt_total_fluxes)
        
        
    @property
    def n_net_fluxes(self):
        
        return len(self._opt_net_fluxes)
        
    
    @property
    def n_free_fluxes(self):
        
        return self.null_space.shape[1]
        
    
    @property
    def opt_objective(self):
        
        return round(self._opt_obj, 3)
        
    
    @property    
    def opt_total_fluxes(self):
        
        return pDict(self._opt_total_fluxes)
    
    
    @property
    def opt_net_fluxes(self):
        
        return pDict(self._opt_net_fluxes)


    @property
    def optimization_successful(self):

        return self.is_success    
    
    
    def chi2_test(self, confidence_level = 0.999):
        '''
        Perform chi square test of the optimal objective.
        SSR < LB of chi square interval sometimes can be also considered as successful estimation.
        
        Parameters
        ----------
        confidence_level: float
            Confidence level, e.g. 0.95 as 95% confidence level.
        '''
        
        _chi2_test(self._opt_obj, self.dof, confidence_level)
        
        
    def plot_normal_probability(self, show_fig = True, output_dir = None):
        '''
        Perform normal probability plot for residuals.
        
        Parameters
        ----------
        show_fig: bool
            Whether to show figure.
        output_dir: str
            Output directory. 
        '''
        
        _normal_probability(self.opt_resids, show_fig, output_dir)
        
    
    def plot_simulated_vs_measured_MDVs(self, show_fig = True, output_dir = None):
        '''
        Plot simulated and measured MDVs.
        
        Parameters
        ----------
        show_fig: bool
            Whether to show figure.
        output_dir: str
            Output directory.
        '''
        
        _simulated_vs_measured_MDVs(
            self.simulated_MDVs, 
            self.measured_MDVs, 
            show_fig, 
            output_dir
        )
        
        
    def plot_simulated_vs_measured_fluxes(self, show_fig = True, output_dir = None):
        '''
        Plot simulated and measured fluxes.
        
        Parameters
        ----------
        show_fig: bool
            Whether to show figure.
        output_dir: str
            Output directory.
        '''
        
        _simulated_vs_measured_fluxes(
            self.simulated_fluxes, 
            self.measured_fluxes, 
            show_fig, 
            output_dir
        )
        
    
    def estimate_confidence_intervals(self, which = 'net', confidence_level = 0.95):
        '''
        Calculate CI of net (total) fluxes using local estimation.
        
        Parameters
        ----------
        which: {"net", "total"}
            * "net" if net fluxes.
            * "total" if total fluxes.
        confidence_level: float
            Confidence level, e.g. 0.95 as 95% confidence level.    
        '''
        
        if which == 'net':
            totalFluxesCov = self.null_space@pinv2(self.hessian)@self.null_space.T
            netFluxesCov = self.transform_matrix@totalFluxesCov@self.transform_matrix.T
            irrRxns = self._opt_net_fluxes.index.intersection(
                self._opt_total_fluxes.index
            ).tolist()
            netFluxesRange = _confidence_intervals_le(
                self._opt_net_fluxes, 
                irrRxns, 
                netFluxesCov, 
                self.dof, 
                confidence_level
            )
            
            return pDict(netFluxesRange)       
        
        elif which == 'total':
            totalFluxesCov = self.null_space@pinv2(self.hessian)@self.null_space.T
            irrRxns = self._opt_total_fluxes.index.tolist()
            totalFluxesRange = _confidence_intervals_le(
                self._opt_total_fluxes, 
                irrRxns, 
                totalFluxesCov, 
                self.dof, 
                confidence_level
            )
            
            return pDict(totalFluxesRange)
        
        else:
            raise ValueError('only "net" and "total" are acceptable for which argument')


    def estimate_contribution_matrix(self, which = 'net'):
        '''
        Calculate contribution matrix of measurement variance to net (total) flux variance.

        Parameters
        ----------
        which: {"net", "total"}
            * "net" if net fluxes.
            * "total" if total fluxes.
        '''

        freeFluxesCov = pinv2(self.hessian)
        expMDVsCov = pinv2(self.exp_MDVs_inv_cov)
        expFluxesCov = pinv2(self.exp_fluxes_inv_cov)

        if which == 'net':
            transMat = self.transform_matrix@self.null_space
            fluxIdx = self._opt_net_fluxes.index
        elif which == 'total':
            transMat = self.null_space
            fluxIdx = self._opt_total_fluxes.index
        else:
            raise ValueError('only "net" and "total" are acceptable for which argument')    

        contribMat_MDV = _contribution_matrix(
            freeFluxesCov, 
            transMat, 
            self.sim_MDVs_der_u, 
            expMDVsCov
        )
        contribMat_MDV = pd.DataFrame(
            contribMat_MDV, 
            index = fluxIdx, 
            columns = self._get_name_of_measurements(self.measured_MDVs)
        )

        contribMat_flux = _contribution_matrix(
            freeFluxesCov, 
            transMat, 
            self.sim_fluxes_der_u, 
            expFluxesCov
        )
        contribMat_flux = pd.DataFrame(
            contribMat_flux, 
            index = fluxIdx,
            columns = self.measured_fluxes.keys()
        )
        
        contribMat = pd.concat((contribMat_MDV, contribMat_flux), axis = 1)

        return contribMat


    def estimate_sensitivity(self, which = 'net'):
        '''
        Calculate sensitivity matrix of estimated net (total) flux w.r.t. measurement changes.
        
        Parameters
        ----------
        which: {"net", "total"}
            * "net" if net fluxes.
            * "total" if total fluxes.
        '''

        freeFluxesCov = pinv2(self.hessian)

        if which == 'net':
            transMat = self.transform_matrix@self.null_space
            fluxIdx = self._opt_net_fluxes.index
        elif which == 'total':
            transMat = self.null_space
            fluxIdx = self._opt_total_fluxes.index
        else:
            raise ValueError('only "net" and "total" are acceptable for which argument')    

        senMat_MDV = _sensitivity(
            freeFluxesCov, 
            transMat, 
            self.sim_MDVs_der_u, 
            self.exp_MDVs_inv_cov
        )
        senMat_MDV = pd.DataFrame(
            senMat_MDV, 
            index = fluxIdx, 
            columns = self._get_name_of_measurements(self.measured_MDVs)
        )
        
        senMat_flux = _sensitivity(
            freeFluxesCov, 
            transMat, 
            self.sim_fluxes_der_u, 
            self.exp_fluxes_inv_cov
        )
        senMat_flux = pd.DataFrame(
            senMat_flux, 
            index = fluxIdx,
            columns = self.measured_fluxes.keys()
        )
        
        senMat = pd.concat((senMat_MDV, senMat_flux), axis = 1)

        return senMat


    @staticmethod
    def _get_name_of_measurements(measured_MDVs):
        '''
        Parameters
        ----------
        measured_MDVs: dict
            EMU ID => [means, sds].
        '''

        names = []
        for emuid in measured_MDVs:
            _, atoms = emuid.split('_')
            for atom in '0'+atoms:
                names.append(f'{emuid}_m{atom}')

        return names        

    
    def __repr__(self):
    
        return f'optimal objective: {self.opt_objective} at\n{self.opt_net_fluxes}'
    
        
    
    
class FitMCResults():
    '''
    Parameters
    ----------
    total_fluxes_set: list of ser
        Set of optimal total fluxes.
    net_fluxes_set: list of ser
        Set of optimal net fluxes.    
    '''
    
    def __init__(self, total_fluxes_set, net_fluxes_set):
        '''
        Parameters
        ----------
        total_fluxes_set: list of ser
            Set of optimal total fluxes.
        net_fluxes_set: list of ser
            Set of optimal net fluxes.    
        '''
        
        self.total_fluxes_set = total_fluxes_set
        self.net_fluxes_set = net_fluxes_set
        
        
    def estimate_confidence_intervals(self, which = 'net', confidence_level = 0.95):
        '''
        Estimate CI from a set of fluxes.
        
        Parameters
        ----------
        which: {"net", "total"}
            * "net" if net fluxes.
            * "total" if total fluxes.
        confidence_level: float
            Confidence level, e.g. 0.95 as 95% confidence level.    
        '''
        
        if which == 'net':
            irrRxns = self.net_fluxes_set[0].index.intersection(
                self.total_fluxes_set[0].index
            ).tolist()
            netFluxesRange = _confidence_intervals_mc(
                self.net_fluxes_set, 
                irrRxns, 
                confidence_level
            )
            return pDict(netFluxesRange)
        
        elif which == 'total':
            irrRxns = self.total_fluxes_set[0].index.tolist()
            totalFluxesRange = _confidence_intervals_mc(
                self.total_fluxes_set, 
                irrRxns, 
                confidence_level
            )
            return pDict(totalFluxesRange)
        
        else:
            raise ValueError('only "net" and "total" are acceptable for which argument')
        
        
    def __repr__(self):
        
        return '%s' % self.estimate_confidence_intervals('net', 0.95)
    
        
        
        
class InstFitResults(FitResults):
    '''
    Parameters
    ----------
    opt_total_fluxes: ser
        Total fluxes at optimal objective.
    opt_net_fluxes: ser
        Net fluxes at optimal objective.
    opt_concs: ser
        Concentrations at optimal objective.
    opt_obj: float
        Optimal value of objective.
    opt_resids: array
        Optimal weighted residuals.
    n_meas: int
        # of measurements.
    n_params: int
        # of parameters.
    sim_inst_MDVs: dict
        EMU ID => {t => simulated MDV}.
    exp_inst_MDVs: dict
        EMU ID => {t => [means, sds]}.
    sim_fluxes: dict
        Flux ID => simulated flux.
    exp_fluxes: dict
        Flux ID => [mean, sd].
    hessian: array
        Hessian matrix at convergence.
    null_space: array
        Null space of stoichiometric matrix.
    transform_matrix: array
        Transform matrix from total fluxes to net fluxes.
    sim_inst_MDVs_der_u: array
        Derivative of simulated MDVs w.r.t. free fluxes.
    sim_fluxes_der_u: array
        Derivative of simualted fluxes w.r.t. free fluxes.
    exp_inst_MDVs_inv_cov: array
        Inversed covariance matrix of measured MDVs.
    exp_fluxes_inv_cov: array
        Inversed covariance matrix of measured fluxes.
    is_success: bool
        Whether the optimization is successful.
    '''
    
    def __init__(
            self, 
            opt_total_fluxes, 
            opt_net_fluxes, 
            opt_concs, 
            opt_obj, 
            opt_resids, 
            n_meas, 
            n_params, 
            sim_inst_MDVs, 
            exp_inst_MDVs, 
            sim_fluxes, 
            exp_fluxes, 
            hessian, 
            null_space, 
            transform_matrix,
            sim_inst_MDVs_der_u, 
            sim_fluxes_der_u, 
            exp_inst_MDVs_inv_cov, 
            exp_fluxes_inv_cov, 
            is_success
    ):
        '''
        Parameters
        ----------
        opt_total_fluxes: ser
            Total fluxes at optimal objective.
        opt_net_fluxes: ser
            Net fluxes at optimal objective.
        opt_concs: ser
            Concentrations at optimal objective.
        opt_obj: float
            Optimal value of objective.
        opt_resids: array
            Optimal weighted residuals.
        n_meas: int
            # of measurements.
        n_params: int
            # of parameters.
        sim_inst_MDVs: dict
            EMU ID => {t => simulated MDV}.
        exp_inst_MDVs: dict
            EMU ID => {t => [means, sds]}.
        sim_fluxes: dict
            Flux ID => simulated flux.
        exp_fluxes: dict
            Flux ID => [mean, sd].
        hessian: array
            Hessian matrix at convergence.
        null_space: array
            Null space of stoichiometric matrix.
        transform_matrix: array
            Transform matrix from total fluxes to net fluxes.
        sim_inst_MDVs_der_u: array
            Derivative of simulated MDVs w.r.t. free fluxes.
        sim_fluxes_der_u: array
            Derivative of simualted fluxes w.r.t. free fluxes.
        exp_inst_MDVs_inv_cov: array
            Inversed covariance matrix of measured MDVs.
        exp_fluxes_inv_cov: array
            Inversed covariance matrix of measured fluxes.
        is_success: bool
            Whether the optimization is successful.
        '''
        
        self._opt_total_fluxes = opt_total_fluxes
        self._opt_net_fluxes = opt_net_fluxes
        self._opt_concs = opt_concs
        self._opt_obj = opt_obj
        self.opt_resids = opt_resids
        self.simulated_inst_MDVs = sim_inst_MDVs
        self.measured_inst_MDVs = exp_inst_MDVs
        self.simulated_fluxes = sim_fluxes
        self.measured_fluxes = exp_fluxes
        self.hessian = hessian
        self.null_space = null_space
        self.transform_matrix = transform_matrix
        self.n_meas = n_meas
        self.n_params = n_params
        self.dof = self.n_meas - self.n_params
        self.sim_inst_MDVs_der_u = sim_inst_MDVs_der_u
        self.sim_fluxes_der_u = sim_fluxes_der_u
        self.exp_inst_MDVs_inv_cov = exp_inst_MDVs_inv_cov
        self.exp_fluxes_inv_cov = exp_fluxes_inv_cov
        self.is_success = is_success
        
    
    @property
    def n_concentrations(self):
        
        return len(self._opt_concs)
        
    
    @property    
    def opt_concentrations(self):
        
        return pDict(self._opt_concs)

        
    def plot_simulated_vs_measured_MDVs(self, show_fig = True, output_dir = None):
        '''
        Plot simulated and measured MDVs.
        
        Parameters
        ----------
        show_fig: bool
            Whether to show figure.
        output_dir: str
            Output directory.
        '''
        
        _simulated_vs_measured_inst_MDVs(
            self.simulated_inst_MDVs, 
            self.measured_inst_MDVs, 
            show_fig, 
            output_dir
        )    
    
    
    def estimate_confidence_intervals(self, which = 'net', confidence_level = 0.95):
        '''
        Calculate CI of fluxes and concentrations using local estimation.
        
        Parameters
        ----------
        which: {"net", "total", "conc"}
            * "net" if net fluxes.
            * "total" if total fluxes.
            * "conc" if concentrations.
        confidence_level: float
            Confidence level, e.g. 0.95 as 95% confidence level.    
        '''
        
        if which == 'net':
            fluxHessian = self.hessian[:self.n_free_fluxes,:self.n_free_fluxes]
            totalFluxesCov = self.null_space@pinv2(fluxHessian)@self.null_space.T
            netFluxesCov = self.transform_matrix@totalFluxesCov@self.transform_matrix.T
            irrItems = self._opt_net_fluxes.index.intersection(
                self._opt_total_fluxes.index
            ).tolist()
            netFluxesRange = _confidence_intervals_le(
                self._opt_net_fluxes, 
                irrItems, 
                netFluxesCov, 
                self.dof, 
                confidence_level
            )
            
            return pDict(netFluxesRange)                                                   
        
        elif which == 'total':
            fluxHessian = self.hessian[:self.n_free_fluxes,:self.n_free_fluxes]
            totalFluxesCov = self.null_space@pinv2(fluxHessian)@self.null_space.T
            irrItems = self._opt_total_fluxes.index.tolist()
            totalFluxesRange = _confidence_intervals_le(
                self._opt_total_fluxes, 
                irrItems, 
                totalFluxesCov, 
                self.dof, 
                confidence_level
            )
            
            return pDict(totalFluxesRange)
        
        elif which == 'conc':
            concHessian = self.hessian[self.n_free_fluxes:, self.n_free_fluxes:]
            concsCov = pinv2(concHessian)
            irrItems = self._opt_concs.index.tolist()
            concsRange = _confidence_intervals_le(
                self._opt_concs, 
                irrItems, 
                concsCov, 
                self.dof, 
                confidence_level
            )
            
            return pDict(concsRange)
            
        else:
            raise ValueError('only "net", "total" and "conc" are acceptable for which argument')
            

    def estimate_contribution_matrix(self, which = 'net'):
        '''
        Calculate contribution matrix of measurement variance to net (total) flux variance.

        Parameters
        ----------
        which: {"net", "total"}
            * "net" if net fluxes.
            * "total" if total fluxes.
        '''

        fluxHessian = self.hessian[:self.n_free_fluxes,:self.n_free_fluxes]
        freeFluxesCov = pinv2(fluxHessian)
        expMDVsCov = pinv2(self.exp_inst_MDVs_inv_cov)
        expFluxesCov = pinv2(self.exp_fluxes_inv_cov)

        if which == 'net':
            transMat = self.transform_matrix@self.null_space
            fluxIdx = self._opt_net_fluxes.index
        elif which == 'total':
            transMat = self.null_space
            fluxIdx = self._opt_total_fluxes.index
        else:
            raise ValueError('only "net" and "total" are acceptable for which argument')    

        contribMat_MDV = _contribution_matrix(
            freeFluxesCov, 
            transMat, 
            self.sim_inst_MDVs_der_u, 
            expMDVsCov
        )
        contribMat_MDV = pd.DataFrame(
            contribMat_MDV, 
            index = fluxIdx, 
            columns = self._get_name_of_measurements(self.measured_inst_MDVs)
        )

        contribMat_flux = _contribution_matrix(
            freeFluxesCov, 
            transMat, 
            self.sim_fluxes_der_u, 
            expFluxesCov
        )
        contribMat_flux = pd.DataFrame(
            contribMat_flux, 
            index = fluxIdx,
            columns = self.measured_fluxes.keys()
        )
        
        contribMat = pd.concat((contribMat_MDV, contribMat_flux), axis = 1)

        return contribMat


    def estimate_sensitivity(self, which = 'net'):
        '''
        Calculate sensitivity matrix of estimated net (total) flux w.r.t. measurement changes.
        
        Parameters
        ----------
        which: {"net", "total"}
            * "net" if net fluxes.
            * "total" if total fluxes.
        '''

        fluxHessian = self.hessian[:self.n_free_fluxes, :self.n_free_fluxes]
        freeFluxesCov = pinv2(fluxHessian)

        if which == 'net':
            transMat = self.transform_matrix@self.null_space
            fluxIdx = self._opt_net_fluxes.index
        elif which == 'total':
            transMat = self.null_space
            fluxIdx = self._opt_total_fluxes.index
        else:
            raise ValueError('only "net" and "total" are acceptable for which argument')    

        senMat_MDV = _sensitivity(
            freeFluxesCov, 
            transMat, 
            self.sim_inst_MDVs_der_u, 
            self.exp_inst_MDVs_inv_cov
        )
        senMat_MDV = pd.DataFrame(
            senMat_MDV, 
            index = fluxIdx, 
            columns = self._get_name_of_measurements(self.measured_inst_MDVs)
        )
        
        senMat_flux = _sensitivity(
            freeFluxesCov, 
            transMat, 
            self.sim_fluxes_der_u, 
            self.exp_fluxes_inv_cov
        )
        senMat_flux = pd.DataFrame(
            senMat_flux, 
            index = fluxIdx,
            columns = self.measured_fluxes.keys()
        )
        
        senMat = pd.concat((senMat_MDV, senMat_flux), axis = 1)

        return senMat
        

    @staticmethod
    def _get_name_of_measurements(measured_MDVs):
        '''
        Parameters
        ----------
        measured_MDVs: dict
            EMU ID => {t => [means, sds]}.
        '''

        names = []
        for emuid in measured_MDVs:
            for t in measured_MDVs[emuid]:
                if t != 0:
                    _, atoms = emuid.split('_')
                    for atom in '0'+atoms:
                        names.append(f'{emuid}_m{atom}_{t}')

        return names




class InstFitMCResults(FitMCResults):
    '''
    Parameters
    ----------
    total_fluxes_set: list of ser
        Set of optimal total fluxes.
    net_fluxes_set: list of ser
        Set of optimal net fluxes.
    concs_set: list of ser
        Set of optimal concentrations.
    '''
    
    def __init__(self, total_fluxes_set, net_fluxes_set, concs_set):
        '''
        Parameters
        ----------
        total_fluxes_set: list of ser
            Set of optimal total fluxes.
        net_fluxes_set: list of ser
            Set of optimal net fluxes.
        concs_set: list of ser
            Set of optimal concentrations.
        '''
        
        super().__init__(total_fluxes_set, net_fluxes_set)
        
        self.concs_set = concs_set


    def estimate_confidence_intervals(self, which = 'net', confidence_level = 0.95):
        '''
        This method estimates CI from a set of fluxes.
        
        Parameters
        ----------
        which: {"net", "total", "conc"}
            * "net" if net fluxes.
            * "total" if total fluxes.
            * "conc" if concentrations.
        confidence_level: float
            Confidence level, e.g. 0.95 as 95% confidence level.    
        '''

        if which == 'net':
            irrItems = self.net_fluxes_set[0].index.intersection(
                self.total_fluxes_set[0].index
            ).tolist()
            netFluxesRange = _confidence_intervals_mc(
                self.net_fluxes_set, 
                irrItems, 
                confidence_level
            )
            return pDict(netFluxesRange)
        
        elif which == 'total':
            irrItems = self.total_fluxes_set[0].index.tolist()
            totalFluxesRange = _confidence_intervals_mc(
                self.total_fluxes_set, 
                irrItems, 
                confidence_level
            )
            return pDict(totalFluxesRange)

        elif which == 'conc':
            irrItems = self.concs_set[0].index.tolist()
            concsRange = _confidence_intervals_mc(
                self.concs_set, 
                irrItems, 
                confidence_level
            )
            return pDict(concsRange)
        
        else:
            raise ValueError('only "net", "total" and "conc" are acceptable for which argument')



'''Define the Simulator class.'''


__author__ = 'Chao Wu'


import re
from collections.abc import Iterable
from functools import partial
from ..core.mdv import MDV
from ..io.inputs import read_preset_values_from_file
from ..io.results import SimResults
from ..utils.utils import Calculator
from ..utils.context import Context
from time import time


class Simulator():
    '''
    Parameters
    ----------
    model: Model
        Freeflux Model.
    '''
    
    def __init__(self, model):
        '''
        Parameters
        ----------
        model: Model
            Freeflux Model.
        '''
        
        self.model = model
        self.calculator = Calculator(self.model)
        self.contexts = []
        
    
    def __enter__(self):
        
        self.contexts.append(Context())

        return self
        
        
    def __exit__(self, type, value, traceback):
        
        context = self.contexts.pop()
        context.undo()
    

    def set_target_EMUs(self, target_emus):
        '''
        Parameters
        ----------
        target_emus: dict
            Metabolite ID => atom NOs or list of atom NOs. Atom NOs can be int list or str, 
            e.g.,, {'Ala': [[1,2,3], [2,3]], 'Ser': '123'}.
        '''
        
        emuids = []
        for metabid, atomNOs in target_emus.items():
            if isinstance(atomNOs, list):
                if any(isinstance(item, Iterable) for item in atomNOs):
                    for atomnos in atomNOs:
                        if not isinstance(atomnos, str):
                            atomnos = ''.join(map(str, atomnos))
                        emuid = metabid+'_'+atomnos
                        emuids.append(emuid)
                        self.model.target_EMUs.append(emuid)
                else:
                    atomNOs = ''.join(map(str, atomNOs))
                    emuid = metabid+'_'+atomNOs
                    emuids.append(emuid)
                    self.model.target_EMUs.append(emuid)
            else:
                emuid = metabid+'_'+atomNOs
                emuids.append(emuid)
                self.model.target_EMUs.append(emuid)
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_target_EMUs, emuids))
        
    
    def _unset_target_EMUs(self, emuids):
        
        for emuid in emuids:
            if emuid in self.model.target_EMUs:
                self.model.target_EMUs.remove(emuid)
    
        
    def set_labeling_strategy(
            self, 
            labeled_substrate, 
            labeling_pattern, 
            percentage, 
            purity,
            label_atom='C'
    ):
        '''
        Use this method for every substrate tracer.
        
        Parameters
        ----------
        labeled_substrate: str
            Metabolite ID.
        labeling_pattern: str or list of str
            Labeling pattern of substrate, '0' for unlabeled atom, '1' for labeled atom, 
            e.g., '100000' for 1-13C glucose. 
            
            List if tracer with multiple labeling patterns are used. 
            
            Natural substrate (with all '0's) don't need to be explicitly set.
            
            If str, labeling_pattern should not be natural substrate.
        percentage: float or list of float
            Molar percentage (in range of [0,1]) of corresponding tracer. 
            Sum of percentage should be <= 1, and the rest will be considered as 
            natural substrate.
            
            List if tracer with multiple labeling patterns are used. 
            
            * If list, len(percentage) should be equal to len(labeling_pattern).
            * If float, labeling_pattern should not be natural substrate.
        purity: float or list of float
            Labeled atom purity (in range of [0,1]) of corresponding tracer.
            
            List if tracer with multiple labeling patterns are used.

            * If list, len(purity) should be equal to len(labeling_pattern).
            * If float, labeling_pattern should not be natural substrate.
        label_atom: str
            Labeled atom, i.e., base atom in the MDV. Currently supports only "H", 
            "C" and "N".
        '''
        
        self.model.labeling_strategy[labeled_substrate] = [labeling_pattern, percentage, purity]
        self.model.label_atom = label_atom
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_labeling_strategy, labeled_substrate))
    
        
    def _unset_labeling_strategy(self, labeled_substrate):
        '''
        Parameters
        ----------
        labeled_substrate: str
            Metabolite ID.
        '''
        
        if labeled_substrate in self.model.labeling_strategy:
            self.model.labeling_strategy.pop(labeled_substrate)
    
    
    def set_flux(self, fluxid, value):
        '''
        Set metabolic flux value.
        
        Parameters
        ----------
        fluxid: str 
            Flux IDs, i.e., reaction ID + '_f' or '_b' for reversible reaction, 
            and reaction ID for irreversible reaction.
        value: float
            Flux value.
        '''
        
        self.model.total_fluxes[fluxid] = value
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_fluxes, fluxid))
        
    
    def set_fluxes_from_file(self, file):
        '''
        Read metabolic flux values from file.
        
        Parameters
        ----------
        file: file path
            tsv or excel file, fields are flux ID and value, 
            flux ID is reaction ID + '_f' or '_b' for reversible reaction, 
            and reaction ID for irreversible reaction.
        '''
        
        fluxes = read_preset_values_from_file(file)
        
        for fluxid, value in fluxes.items():
            self.model.total_fluxes[fluxid] = value
            
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(partial(self._unset_fluxes, fluxes.index.tolist()))
        
                
    def _unset_fluxes(self, fluxids):
        '''
        Parameters
        ----------
        fluxids: str or list of str
            Flux ID(s).
        '''
        
        if not isinstance(fluxids, Iterable):
            fluxids = [fluxids]
        
        for fluxid in fluxids: 
            if fluxid in self.model.total_fluxes:
                self.model.total_fluxes.drop(fluxid, inplace = True)
    
    
    def _decompose_network(self, n_jobs, lump = True):
        '''
        Parameters
        ----------
        n_jobs: int
            # of jobs to run in parallel.
        lump: bool
            Whether to lump linear EMUs.
        '''
        
        if not self.model.target_EMUs:
            raise ValueError('call set_target_EMUs first')
        
        if not self.model.EAMs:
            if n_jobs <= 0:
                raise ValueError('n_jobs should be a positive value')    
            else:
                metabids = []
                atom_nos = []
                for emuid in self.model.target_EMUs:
                    metabid, atomNOs = emuid.split('_')
                    metabids.append(metabid)
                    atom_nos.append(atomNOs)
                
                EAMs = self.model._decompose_network(
                    metabids, 
                    atom_nos, 
                    lump = lump, 
                    n_jobs = n_jobs
                )
                for size, EAM in EAMs.items():
                    self.model.EAMs[size] = EAM
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_decomposition)
            
            
    def _unset_decomposition(self):
        
        self.model.EAMs.clear()
            
    
    def _lambdify_matrix_As_and_Bs(self):
    
        if not self.model.matrix_As or not self.model.matrix_Bs:
            self.calculator._lambdify_matrix_As_and_Bs()
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_matrix_As_and_Bs)        
    
    
    def _unset_matrix_As_and_Bs(self):
        
        self.model.matrix_As.clear()
        self.model.matrix_Bs.clear()
            
    
    def _calculate_substrate_MDVs(self, extra_subs = None):
        '''
        Parameters
        ----------
        extra_subs: str or list of str
            Metabolite ID(s). Additional metabolites that are considered as substrates.
        '''

        if extra_subs is not None and not isinstance(extra_subs, list):
            extra_subs = [extra_subs]

        if not self.model.substrate_MDVs:
            self.calculator._calculate_substrate_MDVs(extra_subs)
        
        if self.contexts:
            context = self.contexts[-1]
            context.add_undo(self._unset_substrate_MDVs)    
    
            
    def _unset_substrate_MDVs(self):
    
        self.model.substrate_MDVs.clear()    
    
    
    def prepare(self, n_jobs = 1):
        '''
        Parameters
        ----------
        n_jobs: int
            If n_jobs > 1, decomposition job will run in parallel.
        '''
        
        self._decompose_network(n_jobs)
        self._lambdify_matrix_As_and_Bs()
        self._calculate_substrate_MDVs()    
    
    
    def _check_dependencies(self):

        if self.model.total_fluxes.empty:
            raise ValueError('call set_flux or set_fluxes_from_file first')
        if not self.model.target_EMUs:
            raise ValueError('call set_target_EMUs first')
        if not self.model.labeling_strategy:
            raise ValueError('call labeling_strategy first')
        
        checklist = [
            not self.model.EAMs, 
            not self.model.substrate_MDVs, 
            not self.model.matrix_As, 
            not self.model.matrix_Bs
        ]
        if any(checklist):
            raise ValueError('call prepare first')


    def simulate(self):
        
        self._check_dependencies()
        
        simMDVs = self.calculator._calculate_MDVs()
        targetMDVs = {emuid: MDV(simMDVs[emuid]) for emuid in self.model.target_EMUs}
        
        return SimResults(targetMDVs)
    